from functools import wraps
from zeep.helpers import serialize_object
from zeep import Client, Settings
from zeep.cache import SqliteCache
from zeep.transports import Transport
import requests
from itertools import product
import json
import logging
import pandas as pd
import multiprocessing
import time
from tqdm import tqdm
from zeep.exceptions import XMLParseError, Fault
import os
import math

from open_mastr.utils import credentials as cred
from open_mastr.soap_api.config import (
    get_filenames,
    setup_logger,
    get_data_version_dir,
    create_data_dir,
)


log = setup_logger()


class MaStRAPI(object):
    """
    Access the Marktstammdatenregister (MaStR) SOAP API via a Python wrapper

    :ref:`Read about <MaStR account and credentials>`
    how to create a user account and a role including a token to access the
    MaStR SOAP API.

    Create an :class:`.MaStRAPI()` instance with your role credentials

    .. code-block:: python

       mastr_api = MaStRAPI(
            user="SOM123456789012",
            key=""koo5eixeiQuoi'w8deighai8ahsh1Ha3eib3coqu7ceeg%ies..."
       )

    Alternatively, leave `user` and `key` empty if user and token are accessible via
    `credentials.cfg`. How to configure this is described :ref:`here <MaStR account and credentials>`.

    .. code-block:: python

        mastr_api = MaStRAPI()

    Now, you can use the MaStR API instance to call `pre-defined SOAP API
    queries
    <https://www.marktstammdatenregister.de/MaStRHilfe/files/webdienst/Funktionen_MaStR_Webdienste_V1.2.39.html>`_
    via the class' methods.
    For example, get a list of units limited to two entries.

    .. code-block:: python

       mastr_api.GetListeAlleEinheiten(limit=2)

    Note, as the example shows, you don't have to pass credentials for calling
    wrapped SOAP queries. This is handled internally.
    """

    def __init__(self, user=None, key=None):
        """
        Parameters
        ----------
        user : str , optional
            MaStR-ID (MaStR-Nummer) for the account that was created on
            https://www.marktstammdatenregister.de
            Typical format: SOM123456789012
        key : str , optional
            Access token of a role (Benutzerrolle). Might look like:
            "koo5eixeiQuoi'w8deighai8ahsh1Ha3eib3coqu7ceeg%ies..."
        """

        # Bind MaStR SOAP API functions as instance methods
        client, client_bind = _mastr_bindings()

        # First, all services of registered service_port (i.e. 'Anlage')
        for n, f in client_bind:
            setattr(self, n, self._mastr_wrapper(f))

        # Second, general functions like 'GetLokaleUhrzeit'
        for n, f in client.service:
            if n == "GetLokaleUhrzeit":
                setattr(self, n, f)
            else:
                setattr(self, n, self._mastr_wrapper(f))

        # Assign MaStR credentials
        if user:
            self._user = user
        else:
            self._user = cred.get_mastr_user()
        if key:
            self._key = key
        else:
            self._key = cred.get_mastr_token(self._user)

    def _mastr_wrapper(self, soap_func):
        """
        Decorates MaStR SOAP API methods with a wrapper automatically passing
        credentials and serializing return value
        """

        @wraps(soap_func)
        def wrapper(*args, **kwargs):
            kwargs.setdefault("apiKey", self._key)
            kwargs.setdefault("marktakteurMastrNummer", self._user)

            # Catch weird MaStR SOAP response
            try:
                response = soap_func(*args, **kwargs)
            except Fault as e:
                # log.warning(f"MaStR SOAP API gives a weird response: {e}. Trying again...")
                time.sleep(1.5)
                try:
                    response = soap_func(*args, **kwargs)
                except Fault as e:
                    msg = (
                        f"MaStR SOAP API still gives a weird response: '{e}'.\n"
                        "Retry failed!"
                    )
                    # log.exception(msg)
                    raise Fault(msg)

            return serialize_object(response, target_cls=dict)

        return wrapper


def _mastr_bindings(
    max_retries=3,
    pool_connections=100,
    pool_maxsize=100,
    timeout=60,
    operation_timeout=600,
    wsdl="https://www.marktstammdatenregister.de/MaStRAPI/wsdl/mastr.wsdl",
    service_name="Marktstammdatenregister",
    service_port="Anlage",
):
    """

    Parameters
    ----------
    max_retries : int
        Maximum number of retries for a request. Parameters is passed to
        requests.adapters.HTTPAdapter
    pool_connections : int
        Number of pool connections. Parameters is passed to
        requests.adapters.HTTPAdapter
    pool_maxsize
        Maximum pool size. Parameters is passed to
        requests.adapters.HTTPAdapter
    timeout : int
        Timeout for loading wsdl sfn xsd documents in seconds. Parameter
        is passed to `zeep.transports.Transport`.
    operation_timeout : int
        Timeout for API requests (GET/POST in underlying requests package)
        in seconds. Parameter is passed to `zeep.transports.Transport`.
    wsdl : str
        Url of wsdl file to be used. Parameters is passed to zeep.Client
    service_name : str
        Service, defined in wsdl file, that is to be used. Parameters is
        passed to zeep.Client.bind
    service_port : str
        Port of service to be used. Parameters is
        passed to zeep.Client.bind

    Returns
    -------
    zeep.Client : The zeep Client
    zeep.Client.bind : ServiceProxy bindings for given :attr:`service_name`
        and :attr:`service_port`
    """

    wsdl = wsdl
    session = requests.Session()
    session.max_redirects = 30
    a = requests.adapters.HTTPAdapter(
        max_retries=max_retries,
        pool_connections=pool_connections,
        pool_maxsize=pool_maxsize,
    )
    session.mount("https://", a)
    transport = Transport(
        cache=SqliteCache(),
        timeout=timeout,
        operation_timeout=operation_timeout,
        session=session,
    )
    settings = Settings(strict=False, xml_huge_tree=True)
    client = Client(wsdl=wsdl, transport=transport, settings=settings)
    client_bind = client.bind(service_name, service_port)

    _mastr_suppress_parsing_errors(["parse-time-second"])

    return client, client_bind


def _mastr_suppress_parsing_errors(which_errors):
    """
    Install logging filters into zeep type parsing modules to suppress

    Arguments
    ---------
    which_errors : [str]
        Names of errors defined in `error_filters` to set up.
        Currently one of ('parse-time-second').

    NOTE
    ----
    zeep and mastr don't seem to agree on the correct time format. Instead of
    suppressing the error, we should fix the parsing error, or they should :).
    """

    class FilterExceptions(logging.Filter):
        def __init__(self, name, klass, msg):
            super().__init__(name)

            self.klass = klass
            self.msg = msg

        def filter(self, record):
            if record.exc_info is None:
                return 1

            kl, inst, tb = record.exc_info
            return 0 if isinstance(inst, self.klass) and inst.args[0] == self.msg else 1

    # Definition of available filters
    error_filters = [
        FilterExceptions("parse-time-second", ValueError, "second must be in 0..59")
    ]

    # Install filters selected by `which_errors`
    zplogger = logging.getLogger("zeep.xsd.types.simple")
    zplogger.filters = [
        f for f in zplogger.filters if not isinstance(f, FilterExceptions)
    ] + [f for f in error_filters if f.name in which_errors]


def flatten_dict(data, serialize_with_json=False):
    """
    Flattens MaStR data dictionary to depth of one

    Parameters
    Parameters
    ----------
    data : list of dict
        Data returned from MaStR-API query

    Returns
    -------
    list of dict
        Flattened data dictionary
    """

    # The rule describes which of the second-level keys are used to replace first-level data
    flatten_rule_replace = {
        "Hausnummer": "Wert",
        "Kraftwerksnummer": "Wert",
        "Weic": "Wert",
        "WeitereBrennstoffe": "Wert",
        "WeitererHauptbrennstoff": "Wert",
        "AnlagenkennzifferAnlagenregister": "Wert",
        "VerhaeltnisErtragsschaetzungReferenzertrag": "Wert",
        "VerhaeltnisReferenzertragErtrag10Jahre": "Wert",
        "VerhaeltnisReferenzertragErtrag15Jahre": "Wert",
        "VerhaeltnisReferenzertragErtrag5Jahre": "Wert",
        "RegistrierungsnummerPvMeldeportal": "Wert",
        "BiogasGaserzeugungskapazitaet": "Wert",
        "BiomethanErstmaligerEinsatz": "Wert",
        "Frist": "Wert",
        "WasserrechtAblaufdatum": "Wert",
    }

    flatten_rule_replace_list = {
        "VerknuepfteEinheit": "MaStRNummer",
        "VerknuepfteEinheiten": "MaStRNummer",
    }

    flatten_rule_serialize = ["Ertuechtigung"]

    flatten_rule_move_up_and_merge = ["Hersteller"]

    flatten_rule_none_if_empty_list = [
        "ArtDerFlaeche",
        "WeitereBrennstoffe",
        "VerknuepfteErzeugungseinheiten",
    ]

    for dic in data:
        # Replacements with second-level values
        for k, v in flatten_rule_replace.items():
            if k in dic.keys():
                dic[k] = dic[k][v]

        # Replacement with second-level value from second-level list
        for k, v in flatten_rule_replace_list.items():
            if k in dic.keys():

                # if data point in 'dic' has one or more VerknuepfteEinheit or
                # VerknuepfteEinheiten in its respective
                # dict, then take first MaStRNummer and insert it into key
                # VerknuepfteEinheiten (flatten dict).
                # Discard all other connected MaStRNummer's of data point.
                if len(dic[k]) != 0:
                    dic[k] = dic[k][0][v]

        # Serilializes dictionary entries with unknown number of sub-entries into JSON string
        # This affects "Ertuechtigung" in extended unit data of hydro
        if serialize_with_json:
            for k in flatten_rule_serialize:
                if k in dic.keys():
                    dic[k] = json.dumps(dic[k], indent=4, sort_keys=True, default=str)

        # Join 'Id' with original key to new column
        # and overwrite original data with 'Wert'
        for k in flatten_rule_move_up_and_merge:
            if k in dic.keys():
                dic.update({k + "Id": dic[k]["Id"]})
                dic.update({k: dic[k]["Wert"]})

        # Avoid empty lists as values
        for k in flatten_rule_none_if_empty_list:
            if k in dic.keys():
                if dic[k] == []:
                    dic[k] = None
                else:
                    dic[k] = ",".join(dic[k])

    return data


def to_csv(df, technology):
    """
    Write joined unit data to CSV file

    Save CSV files to the respective data version directory, see
    :meth:`open_mastr.soap_api.config.get_data_version_dir`.

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe of unit data.
    technology: str
        See list of available technologies in
        :meth:`open_mastr.soap_api.download.MaStRDownload.download_power_plants`.
    """
    data_path = get_data_version_dir()
    filenames = get_filenames()

    csv_file = os.path.join(data_path, filenames["raw"][technology]["joined"])

    df.to_csv(csv_file, index=True, index_label="EinheitMastrNummer", encoding="utf-8")


def _missed_units_to_file(technology, data_type, missed_units):
    """
    Write IDs of missed units to file

    Parameters
    ----------
    technology : str
        Technology, see :meth:`MaStRDownload.download_power_plants`
    data_type : str
        Which type of additional data. Options: 'extended', 'eeg', 'kwk', 'permit'
    missed_units : list
        Unit IDs of missed data
    """

    data_path = get_data_version_dir()
    filenames = get_filenames()
    missed_units_file = os.path.join(
        data_path, filenames["raw"][technology][f"{data_type}_fail"]
    )

    with open(missed_units_file, "w") as f:
        for i, error in missed_units:
            f.write(f"{i},{error}\n")


def _chunksize(length, processes):
    """
    Estimate a useful chunksize for parallel download.

    Depends on the list `length` and the number fo `processes`.

    Parameters
    ----------
    length : int
        Length of list items to be downloaded
    processes : int
        Number of parallel processes

    Returns
    -------
    int
        Chunksize
    """
    if processes > 1:
        chunksize = int(math.ceil(length / (processes * 20)))
    else:
        chunksize = 1

    return chunksize


class _MaStRDownloadFactory(type):
    def __new__(cls, name, bases, dct):
        # Assign factory properties to concrete object
        x = super().__new__(cls, name, bases, dct)

        # Assign mastr_api
        x._mastr_api = MaStRAPI()

        return x


class MaStRDownload(metaclass=_MaStRDownloadFactory):
    """Use the higher level interface for bulk download

    :class:`.MaStRDownload` builds on top of :class:`.MaStRAPI()` and provides
    an interface for easier downloading.
    Use methods documented below to retrieve specific data. On the example of
    data for nuclear power plants, this looks like

    .. code-block:: python

        from open_mastr.soap_api.download import MaStRDownload

        mastr_dl = MaStRDownload()

        for tech in ["nuclear", "hydro", "wind", "solar", "biomass", "combustion", "gsgk"]:
            power_plants = mastr_dl.download_power_plants(tech, limit=10)
            print(power_plants.head())

    .. warning::

        Be careful with increasing `limit`. Typically, your account allows only for 10.000 API
        request per day.

    """

    def __init__(self, parallel_processes=None):
        """

        Parameters
        ----------
        parallel_processes : int or bool, optional
            Specify number of parallel unit data download, respectively
            the number of processes you want to use for downloading.
            For single-process download (avoiding the use of python
            multiprocessing package) choose False.
            Defaults to number of cores (including hyperthreading).
        """

        # Number of parallel processes
        if parallel_processes == "max":
            self.parallel_processes = multiprocessing.cpu_count()
        else:
            self.parallel_processes = parallel_processes

        # Specify which additional data for each unit type is available
        # and which SOAP service has to be used to query it
        self._unit_data_specs = {
            "biomass": {
                "unit_data": "GetEinheitBiomasse",
                "energietraeger": ["Biomasse"],
                "kwk_data": "GetAnlageKwk",
                "eeg_data": "GetAnlageEegBiomasse",
                "permit_data": "GetEinheitGenehmigung",
            },
            "combustion": {
                "unit_data": "GetEinheitVerbrennung",
                "energietraeger": [
                    "Steinkohle",
                    "Braunkohle",
                    "Erdgas",
                    "AndereGase",
                    "Mineraloelprodukte",
                    "NichtBiogenerAbfall",
                    "Waerme",
                ],
                "kwk_data": "GetAnlageKwk",
                "permit_data": "GetEinheitGenehmigung",
            },
            "gsgk": {
                "unit_data": "GetEinheitGeoSolarthermieGrubenKlaerschlammDruckentspannung",
                "energietraeger": [
                    "Geothermie",
                    "Solarthermie",
                    "Grubengas",
                    "Klaerschlamm",
                ],
                "kwk_data": "GetAnlageKwk",
                "eeg_data": "GetAnlageEegGeoSolarthermieGrubenKlaerschlammDruckentspannung",
                "permit_data": "GetEinheitGenehmigung",
            },
            "nuclear": {
                "unit_data": "GetEinheitKernkraft",
                "energietraeger": ["Kernenergie"],
                "permit_data": "GetEinheitGenehmigung",
            },
            "solar": {
                "unit_data": "GetEinheitSolar",
                "energietraeger": ["SolareStrahlungsenergie"],
                "eeg_data": "GetAnlageEegSolar",
                "permit_data": "GetEinheitGenehmigung",
            },
            "wind": {
                "unit_data": "GetEinheitWind",
                "energietraeger": ["Wind"],
                "eeg_data": "GetAnlageEegWind",
                "permit_data": "GetEinheitGenehmigung",
            },
            "hydro": {
                "unit_data": "GetEinheitWasser",
                "energietraeger": ["Wasser"],
                "eeg_data": "GetAnlageEegWasser",
                "permit_data": "GetEinheitGenehmigung",
            },
            "storage": {
                "unit_data": "GetEinheitStromSpeicher",
                "energietraeger": ["Speicher"],
                "eeg_data": "GetAnlageEegSpeicher",
                # todo: additional data request not created for permit, create manually
                "permit_data": "GetEinheitGenehmigung",
            },
            "gas_storage": {
                "unit_data": "GetEinheitGasSpeicher",
                "energietraeger": ["Speicher"],
            },
            # TODO: unsure if energietraeger Ergdas makes sense
            "gas_consumer": {
                "unit_data": "GetEinheitGasVerbraucher",
                "energietraeger": ["Erdgas"],
            },
            "consumer": {
                "unit_data": "GetEinheitStromVerbraucher",
                "energietraeger": ["Strom"],
            },
            "gas_producer": {
                "unit_data": "GetEinheitGasErzeuger",
                "energietraeger": [None],
            },
            "location_elec_generation": "GetLokationStromErzeuger",
            "location_elec_consumption": "GetLokationStromVerbraucher",
            "location_gas_generation": "GetLokationGasErzeuger",
            "location_gas_consumption": "GetLokationGasVerbraucher",
        }

        # Map additional data to primary key via data_fcn
        self._additional_data_primary_key = {
            "extended_unit_data": "EinheitMastrNummer",
            "kwk_unit_data": "KwkMastrNummer",
            "eeg_unit_data": "EegMastrNummer",
            "permit_unit_data": "GenMastrNummer",
            "location_data": "MastrNummer",
        }

        # Check if MaStR credentials are available and otherwise ask
        # for user input
        self._mastr_api._user = cred.check_and_set_mastr_user()
        self._mastr_api._key = cred.check_and_set_mastr_token(self._mastr_api._user)

    def download_power_plants(self, technology, limit=None):
        """
        Download power plant unit data for one technology.

        Based on list with basic information about each unit, subsequently additional
        data is retrieved:

        * Extended unit data
        * EEG data is collected during support of renewable energy installations
          by the Erneuerbare-Energie-Gesetz.
        * KWK data is collected to the support program Kraft-Waerme-Kopplung
        * Permit data is available for some installations (German: Genehmigungsdaten)

        Data is stored in CSV file format in `~/open-MaStR/data/<version>/` by
        default.

        Parameters
        ----------
        technology : str
            Retrieve unit data for one power system unit. Power plants are
            grouped by following technologies:

            * 'nuclear'
            * 'hydro'
            * 'solar'
            * 'wind'
            * 'biomass'
            * 'combustion'
            * 'gsgk'
            * 'storage'

        limit : int
            Maximum number of units to be downloaded. Defaults to :code:`None`.

        Returns
        -------
        pd.DataFrame
            Joined data tables
        """
        # Create data version directory
        create_data_dir()

        # Check requests contingent
        self.daily_contingent()

        # Retrieve basic power plant unit data
        # The return value is casted into a list, because a generator gets returned
        # This was introduced later, after creation of this method
        units = [
            unit
            for sublist in self.basic_unit_data(technology=technology, limit=limit)
            for unit in sublist
        ]

        # Prepare list of unit ID for different additional data (extended, eeg, kwk, permit)
        mastr_ids = [basic["EinheitMastrNummer"] for basic in units]

        # Prepare list of EEG data unit IDs
        if "eeg_data" in self._unit_data_specs[technology].keys():
            eeg_ids = [
                basic["EegMastrNummer"] for basic in units if basic["EegMastrNummer"]
            ]
        else:
            eeg_ids = []

        # Prepare list of KWK data unit IDs
        if "kwk_data" in self._unit_data_specs[technology].keys():
            kwk_ids = [
                basic["KwkMastrNummer"] for basic in units if basic["KwkMastrNummer"]
            ]
        else:
            kwk_ids = []

        # Prepare list of permit data unit IDs
        if "permit_data" in self._unit_data_specs[technology].keys():
            permit_ids = [
                basic["GenMastrNummer"] for basic in units if basic["GenMastrNummer"]
            ]
        else:
            permit_ids = []

        # Download additional data for unit
        extended_data, extended_missed = self.additional_data(
            technology, mastr_ids, "extended_unit_data"
        )
        if eeg_ids:
            eeg_data, eeg_missed = self.additional_data(
                technology, eeg_ids, "eeg_unit_data"
            )
        else:
            eeg_data = eeg_missed = []
        if kwk_ids:
            kwk_data, kwk_missed = self.additional_data(
                technology, kwk_ids, "kwk_unit_data"
            )
        else:
            kwk_data = kwk_missed = []
        if permit_ids:
            permit_data, permit_missed = self.additional_data(
                technology, permit_ids, "permit_unit_data"
            )
        else:
            permit_data = permit_missed = []

        # Retry missed additional unit data
        if extended_missed:
            (
                extended_data_retry,
                extended_missed_retry,
            ) = self._retry_missed_additional_data(
                technology, [_[0] for _ in extended_missed], "extended_unit_data"
            )
            extended_data.extend(extended_data_retry)
            _missed_units_to_file(technology, "extended", extended_missed_retry)
        if eeg_missed:
            eeg_data_retry, eeg_missed_retry = self._retry_missed_additional_data(
                technology, [_[0] for _ in eeg_missed], "eeg_unit_data"
            )
            eeg_data.extend(eeg_data_retry)
            _missed_units_to_file(technology, "eeg", eeg_missed_retry)
        if kwk_missed:
            kwk_data_retry, kwk_missed_retry = self._retry_missed_additional_data(
                technology, [_[0] for _ in kwk_missed], "kwk_unit_data"
            )
            kwk_data.extend(kwk_data_retry)
            _missed_units_to_file(technology, "kwk", kwk_missed_retry)
        if permit_missed:
            permit_data_retry, permit_missed_retry = self._retry_missed_additional_data(
                technology, [_[0] for _ in permit_missed], "permit_unit_data"
            )
            permit_data.extend(permit_data_retry)
            _missed_units_to_file(technology, "permit", permit_missed_retry)

        # Flatten data
        extended_data = flatten_dict(extended_data, serialize_with_json=True)
        eeg_data = flatten_dict(eeg_data, serialize_with_json=True)
        kwk_data = flatten_dict(kwk_data, serialize_with_json=True)
        permit_data = flatten_dict(permit_data, serialize_with_json=True)

        # Join data to a single dataframe
        idx_cols = [
            (units, "EinheitMastrNummer", ""),
            (extended_data, "EinheitMastrNummer", "_unit"),
            (eeg_data, "VerknuepfteEinheit", "_eeg"),
            (kwk_data, "VerknuepfteEinheiten", "_kwk"),
            (permit_data, "VerknuepfteEinheiten", "_permit"),
        ]

        joined_data = pd.DataFrame(idx_cols[0][0]).set_index(idx_cols[0][1])

        for dat, idx_col, suf in idx_cols[1:]:
            # Make sure at least on non-empty dict is in dat
            if any(dat):
                joined_data = joined_data.join(
                    pd.DataFrame(dat).set_index(idx_col), rsuffix=suf
                )

        # Remove duplicates
        joined_data.drop_duplicates(inplace=True)

        to_csv(joined_data, technology)

        return joined_data

    def basic_unit_data(
        self, technology=None, limit=2000, date_from=None, max_retries=3
    ):
        """
        Download basic unit information for one technology.

        Retrieves basic information about units. The number of unit in
        bound to `limit`.

        Parameters
        ----------
        technology : str, optional
            Technology data is requested for. See :meth:`MaStRDownload.download_power_plants`
            for options.
            Data is retrieved using :meth:`MaStRAPI.GetGefilterteListeStromErzeuger`.
            If not given, it defaults to `None`. This implies data for all available
            technologies is retrieved using the web service function
            :meth:`MaStRAPI.GetListeAlleEinheiten`.
        limit : int, optional
            Maximum number of units to download.
            If not provided, data for all units is downloaded.

            .. warning:

               Mind the daily request limit for your MaStR account.

        date_from: `datetime.datetime()`, optional
            If specified, only units with latest change date newer than this are queried.
            Defaults to `None`.
        max_retries: int, optional
            Maximum number of retries in case of errors with the connection to the server.

        Yields
        ------
        list of dict
            A generator of dicts is returned with each dictionary containing
            information about one unit.
        """
        # Split download of basic unit data in chunks of 2000
        # Reason: the API limits retrieval of data to 2000 items
        chunksize = 2000
        chunks_start = list(range(1, limit + 1, chunksize))
        limits = [
            chunksize if (x + chunksize) <= limit else limit - x + 1
            for x in chunks_start
        ]

        # Deal with or w/o technology being specified
        if not technology:
            energietraeger = [None]
        else:
            energietraeger = self._unit_data_specs[technology]["energietraeger"]

        # In case multiple energy carriers (energietraeger) exist for one technology,
        # loop over these and join data to one list
        for et in energietraeger:
            log.info(
                f"Get list of units with basic information for technology {technology} ({et})"
            )
            if et is None:
                query_results = basic_data_download(
                    self._mastr_api,
                    "GetListeAlleEinheiten",
                    "Einheiten",
                    chunks_start,
                    limits,
                    date_from,
                    max_retries,
                    technology,
                    et=et,
                )
            else:
                query_results = basic_data_download(
                    self._mastr_api,
                    "GetGefilterteListeStromErzeuger",
                    "Einheiten",
                    chunks_start,
                    limits,
                    date_from,
                    max_retries,
                    technology,
                    et=et,
                )
            yield from query_results

    def additional_data(self, technology, unit_ids, data_fcn, timeout=10):
        """
        Retrieve addtional informations about units.

        Extended information on units is available. Depending on type, additional data from EEG
        and KWK subsidy program are available.
        Furthermore, for some units, data about permit is retrievable.

        Parameters
        ----------
        technology : str
            Technology, see :meth:`MaStRDownload.download_power_plants`
        unit_ids : list
            Unit identifier for additional data
        data_fcn : str
            Name of method from :class:`MaStRDownload` to be used for querying additional data.
            Choose from

            * "extended_unit_data" (:meth:`~.extended_unit_data`): Extended information
              (i.e. technical, location)
              about a unit. The exact set of information depends on the technology.
            * "eeg_unit_data" (:meth:`~.eeg_unit_data`): Unit Information from
              EEG unit registry. The exact
              set of information depends on the technology.
            * "kwk_unit_data" (:meth:`~.kwk_unit_data`): Unit information from KWK unit registry.
            * "permit_unit_data" (:meth:`~.permit_unit_data`): Information about the permit
              process of a unit.

        timeout: int, optional
            Timeout limit for data retrieval for each unit when using multiprocessing

        Returns
        -------
        tuple of list of dict and tuple
            Returns additional data in dictionaries that are packed into a list. Format

            .. code-block:: python

               return = (
                    [additional_unit_data_dict1, additional_unit_data_dict2, ...],
                    [tuple("SME930865355925", "Reason for failing dowload"), ...]
                    )
        """
        # Prepare a list of unit IDs packed as tuple associated with technology
        prepared_args = list(product(unit_ids, [technology]))

        # Prepare results lists
        data = []
        data_missed = []

        if self.parallel_processes:
            # Estimate a suitable chunksize
            chunksize = _chunksize(len(prepared_args), self.parallel_processes)

            # Open a pool of workers and retrieve data in parallel
            with multiprocessing.Pool(
                processes=self.parallel_processes, maxtasksperchild=1
            ) as pool:

                with tqdm(
                    total=len(prepared_args),
                    desc=f"Downloading {data_fcn} ({technology})",
                    unit="unit",
                ) as pbar:
                    unit_result = pool.imap_unordered(
                        self.__getattribute__(data_fcn), prepared_args, chunksize=1
                    )
                    while True:
                        try:
                            # Try to retrieve data from concurrent processes
                            data_tmp, data_missed_tmp = unit_result.next(
                                timeout=timeout
                            )

                            if not data_tmp:
                                log.debug(
                                    f"Download for additional data for "
                                    f"{data_missed_tmp[0]} ({technology}) failed. "
                                    f"Traceback of caught error:\n{data_missed_tmp[1]}"
                                )
                            data.append(data_tmp)
                            data_missed.append(data_missed_tmp)
                            pbar.update()
                        except StopIteration:
                            # Multiprocessing returns StropIteration when results list gets empty
                            break
                        except multiprocessing.TimeoutError:
                            # If retrieval time exceeds timeout of next(), pass on
                            log.debug(f"Data request for 1 {technology} unit timed out")
        else:
            # Retrieve data in a single process
            for unit_specs in tqdm(
                prepared_args,
                total=len(prepared_args),
                desc=f"Downloading {data_fcn} ({technology})",
                unit="unit",
            ):
                data_tmp, data_missed_tmp = self.__getattribute__(data_fcn)(unit_specs)
                if not data_tmp:
                    log.debug(
                        f"Download for additional data for "
                        f"{data_missed_tmp[0]} ({technology}) failed. "
                        f"Traceback of caught error:\n{data_missed_tmp[1]}"
                    )
                data.append(data_tmp)
                data_missed.append(data_missed_tmp)

        # Remove Nones and empty dicts
        data = [dat for dat in data if dat]
        data_missed = [dat for dat in data_missed if dat]

        # Add units missed due to timeout to data_missed
        units_retrieved = [_[self._additional_data_primary_key[data_fcn]] for _ in data]
        units_missed_timeout = [
            (u, "Timeout")
            for u in unit_ids
            if u not in units_retrieved + [_[0] for _ in data_missed]
        ]
        data_missed = data_missed + units_missed_timeout

        return data, data_missed

    def extended_unit_data(self, unit_specs):
        """
        Download extended data for a unit.

        This extended unit information is provided separately.

        Parameters
        ----------
        unit_specs : tuple
            *EinheitMastrNummer* and technology as tuple that for example looks like

            .. code-block:: python

               tuple("SME930865355925", "hydro")

        Returns
        -------
        dict
            Extended information about unit, if download successful,
            otherwise empty dict
        tuple
            *EinheitMastrNummer* and message the explains why a download failed. Format

            .. code-block:: python

               tuple("SME930865355925", "Reason for failing dowload")
        """

        mastr_id, technology = unit_specs
        try:
            unit_data = self._mastr_api.__getattribute__(
                self._unit_data_specs[technology]["unit_data"]
            )(einheitMastrNummer=mastr_id)
            unit_missed = None
        except (
            XMLParseError,
            Fault,
            requests.exceptions.ConnectionError,
            requests.exceptions.ReadTimeout,
        ) as e:
            # log.exception(
            #     f"Failed to download unit data for {mastr_id} because of SOAP API exception: {e}",
            #     exc_info=False)
            unit_data = {}
            unit_missed = (mastr_id, repr(e))

        return unit_data, unit_missed

    def eeg_unit_data(self, unit_specs):
        """
        Download EEG (Erneuerbare Energien Gesetz) data for a unit.

        Additional data collected during a subsidy program for supporting
        installations of renewable energy power plants.

        Parameters
        ----------
        unit_specs : tuple
            *EegMastrnummer* and technology as tuple that for example looks like

            .. code-block:: python

               tuple("EEG961554380393", "hydro")

        Returns
        -------
        dict
            EEG details about unit, if download successful,
            otherwise empty dict
        tuple
            *EegMastrNummer* and message the explains why a download failed. Format

            .. code-block:: python

               tuple("EEG961554380393", "Reason for failing dowload")
        """
        # TODO: Update docstring to change arguments
        eeg_id, technology = unit_specs
        try:
            eeg_data = self._mastr_api.__getattribute__(
                self._unit_data_specs[technology]["eeg_data"]
            )(eegMastrNummer=eeg_id)
            eeg_missed = None
        except (
            XMLParseError,
            Fault,
            requests.exceptions.ConnectionError,
            requests.exceptions.ReadTimeout,
        ) as e:
            # log.exception(
            #     f"Failed to download eeg data for {eeg_id} because of SOAP API exception: {e}",
            #     exc_info=False)
            eeg_data = {}
            eeg_missed = (eeg_id, repr(e))

        return eeg_data, eeg_missed

    def kwk_unit_data(self, unit_specs):
        """
        Download KWK (Kraft-Wärme-Kopplung) data for a unit.

        Additional data collected during a subsidy program for supporting
        combined heat power plants.

        Parameters
        ----------
        unit_specs : tuple
            *KwkMastrnummer* and technology as tuple that for example looks like

            .. code-block:: python

               tuple("KWK910493229164", "biomass")


        Returns
        -------
        dict
            KWK details about unit, if download successful,
            otherwise empty dict
        tuple
            *KwkMastrNummer* and message the explains why a download failed. Format

            .. code-block:: python

               tuple("KWK910493229164", "Reason for failing dowload")
        """
        kwk_id, technology = unit_specs
        try:
            kwk_data = self._mastr_api.__getattribute__(
                self._unit_data_specs[technology]["kwk_data"]
            )(kwkMastrNummer=kwk_id)
            kwk_missed = None
        except (
            XMLParseError,
            Fault,
            requests.exceptions.ConnectionError,
            requests.exceptions.ReadTimeout,
        ) as e:
            # log.exception(
            #     f"Failed to download unit data for {kwk_id} because of SOAP API exception: {e}",
            #     exc_info=False)
            kwk_data = {}
            kwk_missed = (kwk_id, repr(e))

        return kwk_data, kwk_missed

    def permit_unit_data(self, unit_specs):
        """
        Download permit data for a unit.

        Parameters
        ----------
        unit_specs : tuple
            *GenMastrnummer* and technology as tuple that for example looks like

            .. code-block:: python

               tuple("SGE952474728808", "biomass")


        Returns
        -------
        dict
            Permit details about unit, if download successful,
            otherwise empty dict
        tuple
            *GenMastrNummer* and message the explains why a download failed. Format

            .. code-block:: python

               tuple("GEN952474728808", "Reason for failing dowload")
        """
        permit_id, technology = unit_specs
        try:
            permit_data = self._mastr_api.__getattribute__(
                self._unit_data_specs[technology]["permit_data"]
            )(genMastrNummer=permit_id)
            permit_missed = None
        except (
            XMLParseError,
            Fault,
            requests.exceptions.ConnectionError,
            requests.exceptions.ReadTimeout,
        ) as e:
            # log.exception(
            #     f"Failed to download unit data for {permit_id} "
            #     f"because of SOAP API exception: {e}",
            #     exc_info=False)
            permit_data = {}
            permit_missed = (permit_id, repr(e))

        return permit_data, permit_missed

    def location_data(self, specs):
        """
        Download extended data for a location

        Allows to download additional data for different location types, see *specs*.

        Parameters
        ----------
        specs : tuple
            Location *Mastrnummer* and data_name as tuple that for example looks like

            .. code-block:: python

               tuple("SEL927688371072", "location_elec_generation")


        Returns
        -------
        dict
            Detailed information about a location, if download successful,
            otherwise empty dict
        tuple
            Location *MastrNummer* and message the explains why a download failed. Format

            .. code-block:: python

               tuple("SEL927688371072", "Reason for failing dowload")
        """

        # Unpack tuple argument to two separate variables
        location_id, data_name = specs

        try:
            data = self._mastr_api.__getattribute__(self._unit_data_specs[data_name])(
                lokationMastrNummer=location_id
            )
            missed = None
        except (
            XMLParseError,
            Fault,
            requests.exceptions.ConnectionError,
            requests.exceptions.ReadTimeout,
        ) as e:
            data = {}
            missed = (location_id, repr(e))

        return data, missed

    def _retry_missed_additional_data(
        self, technology, missed_ids, data_fcn, retries=3
    ):
        """
        Retry to download extended data that was missed earlier.

        Tries three times (default) to download data.

        Parameters
        ----------
        technology : str
            Technology, see :meth:`MaStRDownload.download_power_plants`
        missed_ids : list
            Unit identifiers for additional data
        data_fcn : str
            Name of method from :class:`MaStRDownload` to be used for querying additional data
        retries : int
            Number of retries (default: 3).

        Returns
        -------
        tuple of lists
            Queried data and still missed unit IDs are returned as :code:`(data, missed_units)`.
        """

        log.info(
            f"Retrying to download additional data for {len(missed_ids)} "
            f"{technology} units with {retries} retries"
        )

        data = []

        missed_ids_remaining = missed_ids
        for retry in range(1, retries + 1):
            data_tmp, missed_ids_tmp = self.additional_data(
                technology, missed_ids_remaining, data_fcn
            )
            if data_tmp:
                data.extend(data_tmp)
            missed_ids_remaining = [_[0] for _ in missed_ids_tmp]

            if not any(missed_ids_remaining):
                break

        return data, missed_ids_tmp

    def basic_location_data(self, limit=2000, date_from=None, max_retries=3):
        """
        Retrieve basic location data in chunks

        Retrieves data for all types of locations at once using
        :meth:`MaStRAPI.GetListeAlleLokationen`.
        Locations include

        * Electricity generation location (SEL - Stromerzeugungslokation)
        * Electricity consumption location (SVL - Stromverbrauchslokation)
        * Gas generation location (GEL - Gaserzeugungslokation)
        * Gas consumption location (GVL - Gasverbrauchslokation)

        Parameters
        ----------
        limit: int, optional
            Maximum number of locations to download.
            Defaults to 2000.

            .. warning:

               Mind the daily request limit for your MaStR account.

        date_from: :obj:`datetime.datetime`, optional
            If specified, only locations with latest change date newer than this are queried.
            Defaults to :obj:`None`.
        max_retries: int, optional
            Maximum number of retries for each chunk in case of errors with the connection to
            the server.

        Yields
        ------
        generator of generators
            For each chunk a separate generator is returned all wrapped into another generator.
            Access with

            .. code-block:: python

                chunks = mastr_dl.basic_location_data(
                    date_from=datetime.datetime(2020, 11, 7, 0, 0, 0), limit=2010
                    )

                for chunk in chunks:
                    for location in chunk:
                        print(location) # prints out one dict per location one after another
        """
        # Prepare indices for chunked data retrieval
        chunksize = 2000
        chunks_start = list(range(1, limit + 1, chunksize))
        limits = [
            chunksize if (x + chunksize) <= limit else limit - x + 1
            for x in chunks_start
        ]

        # Retrieve query results and yield them
        query_results = basic_data_download(
            self._mastr_api,
            "GetListeAlleLokationen",
            "Lokationen",
            chunks_start,
            limits,
            date_from,
            max_retries,
        )
        yield from query_results

    def daily_contingent(self):
        contingent = self._mastr_api.GetAktuellerStandTageskontingent()
        log.info(
            f"Daily requests contigent: "
            f"{contingent['AktuellerStandTageskontingent']} "
            f"/ {contingent['AktuellesLimitTageskontingent']}"
        )


def basic_data_download(
    mastr_api,
    fcn_name,
    category,
    chunks_start,
    limits,
    date_from,
    max_retries,
    technology=None,
    et=None,
):
    """
    Helper function for downloading basic data with MaStR list query

    Helps to query data from list-returning functions like GetListeAlleEinheiten.
    Automatically

    * respects limit of 2.000 rows returned by MaStR list functions
    * stops, if no further data is available
    * nicely integrates dynamic update of tqdm progress bar

    Parameters
    ----------
    mastr_api: :class:`MaStRAPI`
        MaStR API wrapper
    fcn_name: str
        Name of list-returning download function
    category: str
        Either "Einheiten" or "Lokationen"
    chunks_start: list of int
        Start index for each chunk. MaStR data is internally index by an integer index that can
        be used to start querying data from a certain position. Here, it is used to
        iterate over the available data.
    limits: list of int
        Limit of queried row for each chunk.
    date_from: datetime.datetime
        Date for querying only newer data than this date
    max_retries: int
        Number of maximum retries for each chunk
    technology: str, optional
        Choose a subset from available technologies. Only relevant if category="Einheiten".
        Defaults to all technologies.
    et: str
        Energietraeger of a technology. Some technologies are subdivided into a list of
        energietraeger. Only relevant if category="Einheiten". Defaults to None.

    Yields
    ------
    list
        Basic unit or location information. Depends on category.
    """

    # Construct description string
    description = f"Get basic {category} data information"
    if technology:
        description += f" for technology {technology}"
    if et:
        description += f" ({et})"

    pbar = tqdm(desc=description, unit=" units")

    # Iterate over chunks and download data
    # Results are first collected per 'et' (units_tech) for properly
    # displaying download progress.
    # Later, all units of a single technology are collected in 'units'
    for chunk_start, limit_iter in zip(chunks_start, limits):
        # Use a retry loop to retry on connection errors
        for try_number in range(max_retries + 1):
            try:
                if et is None:
                    response = getattr(mastr_api, fcn_name)(
                        startAb=chunk_start, limit=limit_iter, datumAb=date_from
                    )
                else:
                    response = getattr(mastr_api, fcn_name)(
                        energietraeger=et,
                        startAb=chunk_start,
                        limit=limit_iter,
                        datumAb=date_from,
                    )

            except (
                requests.exceptions.ConnectionError,
                Fault,
                requests.exceptions.ReadTimeout,
            ) as e:
                try_number += 1
                log.debug(
                    f"MaStR SOAP API does not respond properly: {e}. Retry {try_number}"
                )
                time.sleep(5)
            else:
                # If it does run into the except clause, break out of the for loop
                # This also means query was successful
                units_tech = response[category]
                yield units_tech
                pbar.update(len(units_tech))
                break
        else:
            log.error(
                f"Finally failed to download data."
                f"Basic unit data of index {chunk_start} to "
                f"{chunk_start + limit_iter - 1} will be missing."
            )
            # TODO: this has potential risk! Please change
            # If the download continuously fails on the last chunk, this query will run forever
            response = {"Ergebniscode": "OkWeitereDatenVorhanden"}

        # Stop querying more data, if no further data available
        if response["Ergebniscode"] == "OkWeitereDatenVorhanden":
            continue
        else:
            # Update progress bar and move on with next et or technology
            pbar.total = pbar.n
            pbar.refresh()
            pbar.close()
            break

    # Make sure progress bar is closed properly
    pbar.close()


if __name__ == "__main__":
    pass

# TODO: Pass through kargs to _unit_data() that are possible 
# to use with GetGefilterteListeStromErzeuger() and mention in docs



# system catalog is the mapping for the entries within the two columns
# Marktfunktionen und Lokationstyp (entry 1 is mapped to Stromnetzbetreiber
# in the column Marktfunktionen)

system_catalog = {
    "Marktfunktion": {
        1: "Stromnetzbetreiber",
        2: "Anlagenbetreiber",
        3: "Akteur im Strommarkt",
        4: "Organisierter Marktplatz",
        5: "Behörde, Verband, Institution",
        6: "Sonstiger Marktakteur",
        7: "Bundesnetzagentur",
        8: "Gasnetzbetreiber",
        9: "Akteur im Gasmarkt",
        10: "Supportpartner",
    },
    "Lokationtyp": {
        1: "Stromerzeugungslokation",
        2: "Stromverbrauchslokation",
        3: "Gaserzeugungslokation",
        4: "Gasverbrauchslokation",
    },
}

# columns to replace lists all columns where the entries have
# to be replaced according to the tables katalogwerte and katalogeinträge
# from the bulk download of the MaStR

columns_replace_list = [
    # anlageneegsolar
    "AnlageBetriebsstatus",
    # anlageneegspeicher
    # anlagenstromspeicher
    # einheitensolar
    "Land",
    "Bundesland",
    "EinheitSystemstatus",
    "EinheitBetriebsstatus",
    "Energietraeger",
    "Einspeisungsart",
    "GemeinsamerWechselrichterMitSpeicher",
    "Lage",
    "Leistungsbegrenzung",
    "Hauptausrichtung",
    "HauptausrichtungNeigungswinkel",
    "Nutzungsbereich",
    "Nebenausrichtung",
    "NebenausrichtungNeigungswinkel",
    # einheitenstromspeicher
    "AcDcKoppelung",
    "Batterietechnologie",
    "Technologie",
    "Pumpspeichertechnologie",
    "Einsatzort",
    # lokationen
    # marktakteure
    "Personenart",
    "Rechtsform",
    "HauptwirtdschaftszweigAbteilung",
    "HauptwirtdschaftszweigGruppe",
    "HauptwirtdschaftszweigAbschnitt",
    "Registergericht",
    "LandAnZustelladresse",
    # netzanschlusspunkte
    "Gasqualitaet",
    "Spannungsebene",
    # anlageneegbiomasse
    # anlageneeggeosolarthermiegrubenklaerschlammdruckentspannung
    # anlageneegwasser
    # anlageneegwind
    # anlagengasspeicher
    # anlagenkwk
    # bilanzierungsgebiete
    # einheitenbiomasse
    "Hauptbrennstoff",
    "Biomasseart",
    # einheitengaserzeuger
    # einheitengasspeicher
    "Speicherart",
    # einheitengasverbraucher
    # einheitengenehmigung
    "Art",
    # einheitengeosolarthermiegrubenklaerschlammdruckentspannung
    # einheitenkernkraft
    # einheitenstromverbraucher
    "ArtAbschaltbareLast",
    # einheitentypen
    # einheitenverbrennung
    "WeitererHauptbrennstoff",
    "ArtDerStilllegung",
    # einheitenwasser
    "ArtDesZuflusses",
    # marktrollen
    # netze
    "Sparte",
    # einheitenwind
    "Lage",
    "Hersteller",
    "Seelage",
    "ClusterNordsee",
    "ClusterOstsee",
]



import datetime
import json
import os
import pandas as pd
import pathlib
import pynodo
import yaml

from open_mastr.soap_api.config import get_filenames, get_data_version_dir, get_power_unit_types
from open_mastr.soap_api.metadata.create import datapackage_meta_json
from open_mastr.utils.credentials import get_zenodo_token


dtypes = {
    "Postleitzahl": str,
    "Gemeindeschluessel": str,
    "Bruttoleistung": float
}


def filter(df):
    """
    Filter unit data based on characters of index labels

    Data is filtered by unit category. Only directly in MaStR entered data is considered; migrated data is filtered out.
    # TODO: Link to data description about migrated, duplicates update process

    Parameters
    ----------
    df. pandas.DataFrame
        Unit data

    Returns
    -------
    pandas.DataFrame
        Filtered dataframe without migrated data
    """
    return df[df.index.str.match("^[S,G][E,S,V][E,L]")]


def convert_datetime(row):
    """
    String to datetime conversion considering different string formats

    Parameters
    ----------
    row: str
        String to convert or pandas.DataFrame cell

    Returns
    -------
    pd.Timestamp
        Pandas timestamp is returned which bases on datetime
    """
    try:
        datetime_date = pd.to_datetime(row, format='%Y-%m-%d %H:%M:%S.%f%z')
    except ValueError:
        datetime_date = pd.to_datetime(row, format='%Y-%m-%d %H:%M:%S', errors='coerce')
    return datetime_date


def read_csv_data(data_stage):
    """Read raw open-MaStR data from project home dir

    Parameters
    ----------
    data_stage: str
        Select which data is loaded from CSV files

        * "raw": Raw MaStR data is read
        * "cleaned": Cleaned MaStR data is read
    
    """

    data_dir = get_data_version_dir()
    filenames = get_filenames()
    power_unit_types = get_power_unit_types()

    data = {}
    for technology, filename in filenames[data_stage].items():
        if technology in power_unit_types:
            if data_stage == "raw" :
                filename = filename["joined"]
            data_file = os.path.join(data_dir, filename)
            if os.path.isfile(data_file):
                data[technology] = pd.read_csv(
                        data_file,
                        parse_dates=["Inbetriebnahmedatum",
                            "DatumLetzteAktualisierung"],
                        index_col="EinheitMastrNummer",
                        sep=",",
                        date_parser=convert_datetime,
                        dtype=dtypes
                        )
            else:
                pass

    return data


def save_cleaned_data(data):
    """
    Save cleaned open-MaStR to CSV files in desired location in project home dir

    Parameters
    ----------
    data: dict of pandas.DataFrame
        Cleaned open-MaStR data in a dict keyed by technology
    """
    data_dir = get_data_version_dir()
    filenames = get_filenames()

    newest_date = datetime.datetime(1900, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)

    # Iterate through dict and save individual dataframes to CSV
    for tech, dat in data.items():
        data_file = os.path.join(data_dir, filenames["cleaned"][tech])
        dat.to_csv(data_file, index=True, index_label="EinheitMastrNummer", encoding='utf-8')

        if dat["DatumLetzteAktualisierung"].max() > newest_date:
            newest_date = dat["DatumLetzteAktualisierung"].max()

    # Save metadata along with data
    metadata_file = os.path.join(data_dir, filenames["metadata"])
    metadata = datapackage_meta_json(newest_date, data.keys(), data=["raw", "cleaned"], json_serialize=False)

    with open(metadata_file, 'w', encoding='utf-8') as f:
        json.dump(metadata, f, ensure_ascii=False, indent=4)


def zenodo_upload(data_stages=["raw", "cleaned", "postprocessed"], zenodo_token=None, sandbox=True):
    """
    Upload MaStR data to a new Zenodo deposit

    Metadata stored in `datapackage.json` is also uploaded.

    Parameters
    ----------
    data_stages: list, optional
        Data is available at three stages: 'raw', 'cleaned' and 'processed'. The data upload only includes data stages
        specified here. Defaults to all data.
    zenodo_token: str, optional
        Uploading to Zenodo requires authentication. Either provide your token here or store it in the credentials.cfg
        file. See also in :ref:`Zenodo token`.
    sandbox: bool
        Flag to toggle between `Zenodo production <https://www.zenodo.org/>`_ and
        `sandbox instance <https://sandbox.zenodo.org>`_. If True, the sandbox instance will be used.
        Specify False, to upload a publication ready dataset of the production instance.
    """

    data_dir = get_data_version_dir()
    filenames = get_filenames()
    metadata_file = os.path.join(data_dir, filenames["metadata"])
    power_unit_types = get_power_unit_types()

    if not zenodo_token:
        # If no token for Zenodo is given, read from credentials.yml
        zenodo_token = get_zenodo_token()
    # If neither a Zenodo token exists in config, raise Error
    if not zenodo_token:
        raise ValueError("No Zenodo token provided. Can't upload.")


    # Prepare metadata for Zenodo deposit
    with open(metadata_file, "r") as read_file:
        metadata = json.load(read_file)
    zenodo_required_metadata = {
        "upload_type": "dataset",
        "publication_date": metadata["created"].split(" ")[0],
        "creators": [{
            "name": contributor["title"],
            "affiliation": contributor["organization"]
        }
            for contributor in metadata["contributors"]],
        "access_right": "open",
        "license": "other-at",
        "title": metadata["title"],
        "description": metadata["description"],
        "version": metadata["version"],
    }

    # Create deposit
    zen = pynodo.Depositions(access_token=zenodo_token, sandbox=sandbox)
    deposit = zen.create(data={"metadata": zenodo_required_metadata})

    # Access deposit files for uploading files
    zen_files = pynodo.DepositionFiles(
        deposition=deposit.id,
        access_token=zenodo_token,
        sandbox=True,
    )

    # Upload LICENSE file
    zen_files.upload(os.path.join(pathlib.Path(__file__).parent.absolute(), "soap_api", "metadata", "LICENSE"))

    # Upload actual data
    for data_stage in data_stages:
        for tech, file_specs in filenames[data_stage].items():
            if data_stage == "raw" and tech in power_unit_types:
                zen_files.upload(os.path.join(data_dir, file_specs["joined"]))
            else:
                zen_files.upload(os.path.join(data_dir, file_specs))

    # Upload datapackage.json
    zen_files.upload(os.path.join(data_dir, filenames["metadata"]))




import datetime
import json
import os
import pandas as pd
from sqlalchemy.orm import sessionmaker, Query
from sqlalchemy import and_, create_engine, func
from sqlalchemy.sql import exists, insert, literal_column
import shlex
import subprocess
from open_mastr.settings import DB_ENGINE
from datetime import date

from open_mastr.soap_api.config import (
    setup_logger,
    create_data_dir,
    get_filenames,
    get_data_version_dir,
    column_renaming,
)
from open_mastr.soap_api.download import MaStRDownload, flatten_dict, to_csv
from open_mastr import orm
from open_mastr.soap_api.metadata.create import datapackage_meta_json
from open_mastr.utils.helpers import session_scope

log = setup_logger()


class MaStRMirror:
    """
    Mirror the Marktstammdatenregister database and keep it up-to-date

    A PostgreSQL database is used to mirror the MaStR database. It builds on functionality for bulk data download
    provided by :class:`open_mastr.soap_api.download.MaStRDownload`.

    A rough overview is given by the following schema on the example of wind power units.

    .. figure:: /images/MaStR_Mirror.svg
       :width: 70%
       :align: center

    Initially, basic unit data gets backfilled with :meth:`~.backfill_basic` (downloads basic unit data for 2,000
    units of type 'solar').

    .. code-block:: python

       from open_mastr.soap_api.prototype_mastr_reflected import MaStRMirror

       mastr_mirror = MaStRMirror()
       mastr_mirror.backfill_basic("solar", limit=2000)

    Based on this, requests for
    additional data are created. This happens during backfilling basic data. But it is also possible to (re-)create
    requests for remaining additional data using :meth:`~.create_additional_data_requests`.

    .. code-block:: python

       mastr_mirror.create_additional_data_requests("solar")

    Additional unit data, in the case of wind power this is extended data, EEG data and permit data, can be
    retrieved subsequently by :meth:`~.retrieve_additional_data`.

    .. code-block:: python

       mastr_mirror.retrieve_additional_data("solar", ["unit_data"])


    The data can be joined to one table for each technology and exported to CSV files using :meth:`~.to_csv`.

    Also consider to use :meth:`~.dump` and :meth:`~.restore` for specific purposes.

    """

    def __init__(
        self,
        empty_schema=False,
        restore_dump=None,
        DB_ENGINE=None,
        parallel_processes=None,
    ):
        """
        Parameters
        ----------
        empty_schema: boolean
            Remove all data from the MaStR mirror database. Deletes the entire schema and recreates it
            including all tables. Be careful!
            Defaults to False **not** deleting anything.
        restore_dump: str or path-like, optional
            Save path of SQL dump file including filename. The database is restored from the SQL dump.
            Defaults to `None` which means nothing gets restored.
            Should be used in combination with `empty_schema=True`.
        DB_ENGINE: str
            Spin up a PostgreSQL database in a docker container. This calls :meth:`~.setup_docker` during instantiation.
        parallel_processes: int
            Number of parallel processes used to download additional data.
            Defaults to `None`.
        """

        # Spin up database container
        if DB_ENGINE == "docker":
            self.setup_docker()

        # Associate downloader
        self.mastr_dl = MaStRDownload(parallel_processes=parallel_processes)

        # Restore database from a dump
        if restore_dump:
            self.restore(restore_dump)

        # Map technologies on ORMs
        self.orm_map = {
            "wind": {
                "unit_data": "WindExtended",
                "eeg_data": "WindEeg",
                "permit_data": "Permit",
            },
            "solar": {
                "unit_data": "SolarExtended",
                "eeg_data": "SolarEeg",
                "permit_data": "Permit",
            },
            "biomass": {
                "unit_data": "BiomassExtended",
                "eeg_data": "BiomassEeg",
                "kwk_data": "Kwk",
                "permit_data": "Permit",
            },
            "combustion": {
                "unit_data": "CombustionExtended",
                "kwk_data": "Kwk",
                "permit_data": "Permit",
            },
            "gsgk": {
                "unit_data": "GsgkExtended",
                "eeg_data": "GsgkEeg",
                "kwk_data": "Kwk",
                "permit_data": "Permit",
            },
            "hydro": {
                "unit_data": "HydroExtended",
                "eeg_data": "HydroEeg",
                "permit_data": "Permit",
            },
            "nuclear": {"unit_data": "NuclearExtended", "permit_data": "Permit"},
            "storage": {
                "unit_data": "StorageExtended",
                "eeg_data": "StorageEeg",
                "permit_data": "Permit",
            },
        }

        # Map technology and MaStR unit type
        # Map technologies on ORMs
        self.unit_type_map = {
            "Windeinheit": "wind",
            "Solareinheit": "solar",
            "Biomasse": "biomass",
            "Wasser": "hydro",
            "Geothermie": "gsgk",
            "Verbrennung": "combustion",
            "Kernenergie": "nuclear",
            "Stromspeichereinheit": "storage",
            "Gasspeichereinheit": "gas_storage",
            "Gasverbrauchseinheit": "gas_consumer",
            "Stromverbrauchseinheit": "consumer",
            "Gaserzeugungseinheit": "gas_producer",
            "Stromerzeugungslokation": "location_elec_generation",
            "Stromverbrauchslokation": "location_elec_consumption",
            "Gaserzeugungslokation": "location_gas_generation",
            "Gasverbrauchslokation": "location_gas_consumption",
        }
        self.unit_type_map_reversed = {v: k for k, v in self.unit_type_map.items()}

    def setup_docker(self):
        """Initialize a PostgreSQL database with docker-compose"""

        conf_file_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "..", "..")
        )
        print(conf_file_path)
        subprocess.run(
            ["docker-compose", "up", "-d"],
            cwd=conf_file_path,
        )

    def backfill_basic(self, technology=None, date=None, limit=None):
        """Backfill basic unit data.

        Fill database table 'basic_units' with data. It allows specification of which data should be retrieved via
        the described parameter options.

        Under the hood, :meth:`open_mastr.soap_api.download.MaStRDownload.basic_unit_data` is used.

        Parameters
        ----------
        technology: str or list
            Specify technologies for which data should be backfilled.

            * 'solar' (`str`): Backfill data for a single technology.
            * ['solar', 'wind'] (`list`):  Backfill data for multiple technologies given in a list.
            * `None`: Backfill data for all technologies

            Defaults to `None` which is passed to :meth:`open_mastr.soap_api.download.MaStRDownload.basic_unit_data`.
        date: None, :class:`datetime.datetime`, str
            Specify backfill date from which on data is retrieved

            Only data with modification time stamp greater that `date` is retrieved.

            * `datetime.datetime(2020, 11, 27)`: Retrieve data which is is newer than this time stamp
            * 'latest': Retrieve data which is newer than the newest data already in the table.
              It is aware of a different 'latest date' for each technology. Hence, it works in combination with
              `technology=None` and `technology=["wind", "solar"]` for example.

              .. warning::

                 Don't use 'latest' in combination with `limit`. This might lead to unexpected results.
            * `None`: Complete backfill

            Defaults to `None`.
        limit: int
            Maximum number of units.
            Defaults to `None` which means no limit is set and all available data is queried. Use with care!
        """

        # Create list of technologies to backfill
        if isinstance(technology, str):
            technology_list = [technology]
        elif technology == None:
            technology_list = [None]
        elif isinstance(technology, list):
            technology_list = technology

        # Set limit to a number >> number of units of technology with most units
        if limit is None:
            limit = 10 ** 8

        if date == "latest":
            dates = []
            for tech in technology_list:
                if tech:
                    # In case technologies are specified, latest data date gets queried per technology
                    with session_scope() as session:
                        newest_date = (
                            session.query(orm.BasicUnit.DatumLetzteAktualisierung)
                            .filter(
                                orm.BasicUnit.Einheittyp
                                == self.unit_type_map_reversed[tech]
                            )
                            .order_by(orm.BasicUnit.DatumLetzteAktualisierung.desc())
                            .first()
                        )
                else:
                    # If technologies aren't defined ([None]) latest date per technology is queried in query
                    # This also leads that the remainder of the loop body is skipped
                    with session_scope() as session:
                        subquery = session.query(
                            orm.BasicUnit.Einheittyp,
                            func.max(orm.BasicUnit.DatumLetzteAktualisierung).label(
                                "maxdate"
                            ),
                        ).group_by(orm.BasicUnit.Einheittyp)
                        dates = [s[1] for s in subquery]
                        technology_list = [self.unit_type_map[s[0]] for s in subquery]
                        # Break the for loop over technology here, because we write technology_list and dates at once
                        break

                # Add date to dates list
                if newest_date:
                    dates.append(newest_date[0])
                # Cover the case where no data is in the database and latest is still used
                else:
                    dates.append(None)
        else:
            dates = [date] * len(technology_list)

        # Retrieve data for each technology separately
        for tech, date in zip(technology_list, dates):
            log.info(f"Backfill data for technology {tech}")

            # Catch weird MaStR SOAP response
            basic_units = self.mastr_dl.basic_unit_data(tech, limit, date_from=date)

            with session_scope() as session:

                # Insert basic data into database
                log.info(
                    "Insert basic unit data into DB and submit additional data requests"
                )
                for basic_units_chunk in basic_units:
                    # Make sure that no duplicates get inserted into database (would result in an error)
                    # Only new data gets inserted or data with newer modification date gets updated

                    # Remove duplicates returned from API
                    basic_units_chunk_unique = [
                        unit
                        for n, unit in enumerate(basic_units_chunk)
                        if unit["EinheitMastrNummer"]
                        not in [
                            _["EinheitMastrNummer"] for _ in basic_units_chunk[n + 1 :]
                        ]
                    ]
                    basic_units_chunk_unique_ids = [
                        _["EinheitMastrNummer"] for _ in basic_units_chunk_unique
                    ]

                    # Find units that are already in the DB
                    common_ids = [
                        _.EinheitMastrNummer
                        for _ in session.query(orm.BasicUnit.EinheitMastrNummer).filter(
                            orm.BasicUnit.EinheitMastrNummer.in_(
                                basic_units_chunk_unique_ids
                            )
                        )
                    ]

                    # Create instances for new data and for updated data
                    insert = []
                    updated = []
                    for unit in basic_units_chunk_unique:

                        # Rename the typo in column DatumLetzeAktualisierung
                        if "DatumLetzteAktualisierung" not in unit.keys():
                            unit["DatumLetzteAktualisierung"] = unit.pop(
                                "DatumLetzeAktualisierung", None
                            )

                        # In case data for the unit already exists, only update if new data is newer
                        if unit["EinheitMastrNummer"] in common_ids:
                            if session.query(
                                exists().where(
                                    and_(
                                        orm.BasicUnit.EinheitMastrNummer
                                        == unit["EinheitMastrNummer"],
                                        orm.BasicUnit.DatumLetzteAktualisierung
                                        < unit["DatumLetzteAktualisierung"],
                                    )
                                )
                            ).scalar():
                                updated.append(unit)
                                session.merge(orm.BasicUnit(**unit))
                        # In case of new data, just insert
                        else:
                            insert.append(unit)
                    session.bulk_save_objects([orm.BasicUnit(**u) for u in insert])
                    inserted_and_updated = insert + updated

                    # Submit additional data requests
                    extended_data = []
                    eeg_data = []
                    kwk_data = []
                    permit_data = []

                    for basic_unit in inserted_and_updated:
                        # Extended unit data
                        extended_data.append(
                            {
                                "EinheitMastrNummer": basic_unit["EinheitMastrNummer"],
                                "additional_data_id": basic_unit["EinheitMastrNummer"],
                                "technology": self.unit_type_map[
                                    basic_unit["Einheittyp"]
                                ],
                                "data_type": "unit_data",
                                "request_date": datetime.datetime.now(
                                    tz=datetime.timezone.utc
                                ),
                            }
                        )

                        # EEG unit data
                        if basic_unit["EegMastrNummer"]:
                            eeg_data.append(
                                {
                                    "EinheitMastrNummer": basic_unit[
                                        "EinheitMastrNummer"
                                    ],
                                    "additional_data_id": basic_unit["EegMastrNummer"],
                                    "technology": self.unit_type_map[
                                        basic_unit["Einheittyp"]
                                    ],
                                    "data_type": "eeg_data",
                                    "request_date": datetime.datetime.now(
                                        tz=datetime.timezone.utc
                                    ),
                                }
                            )

                        # KWK unit data
                        if basic_unit["KwkMastrNummer"]:
                            kwk_data.append(
                                {
                                    "EinheitMastrNummer": basic_unit[
                                        "EinheitMastrNummer"
                                    ],
                                    "additional_data_id": basic_unit["KwkMastrNummer"],
                                    "technology": self.unit_type_map[
                                        basic_unit["Einheittyp"]
                                    ],
                                    "data_type": "kwk_data",
                                    "request_date": datetime.datetime.now(
                                        tz=datetime.timezone.utc
                                    ),
                                }
                            )

                        # Permit unit data
                        if basic_unit["GenMastrNummer"]:
                            permit_data.append(
                                {
                                    "EinheitMastrNummer": basic_unit[
                                        "EinheitMastrNummer"
                                    ],
                                    "additional_data_id": basic_unit["GenMastrNummer"],
                                    "technology": self.unit_type_map[
                                        basic_unit["Einheittyp"]
                                    ],
                                    "data_type": "permit_data",
                                    "request_date": datetime.datetime.now(
                                        tz=datetime.timezone.utc
                                    ),
                                }
                            )

                # Delete old entries for additional data requests
                additional_data_table = orm.AdditionalDataRequested.__table__
                ids_to_delete = [_["EinheitMastrNummer"] for _ in inserted_and_updated]
                session.execute(
                    additional_data_table.delete()
                    .where(
                        additional_data_table.c.EinheitMastrNummer.in_(ids_to_delete)
                    )
                    .where(additional_data_table.c.technology == "wind")
                    .where(
                        additional_data_table.c.request_date
                        < datetime.datetime.now(tz=datetime.timezone.utc)
                    )
                )

                # Flush delete statements to database
                session.commit()

                # Insert new requests for additional data
                session.bulk_insert_mappings(orm.AdditionalDataRequested, extended_data)
                session.bulk_insert_mappings(orm.AdditionalDataRequested, eeg_data)
                session.bulk_insert_mappings(orm.AdditionalDataRequested, kwk_data)
                session.bulk_insert_mappings(orm.AdditionalDataRequested, permit_data)

            log.info("Backfill successfully finished")

    def backfill_locations_basic(
        self, limit=None, date=None, delete_additional_data_requests=True
    ):
        """
        Backfill basic location data.

        Fill database table 'locations_basic' with data. It allows specification of which data should be retrieved via
        the described parameter options.

        Under the hood, :meth:`open_mastr.soap_api.download.MaStRDownload.basic_location_data` is used.

        Parameters
        ----------
        date: None, :class:`datetime.datetime`, str
            Specify backfill date from which on data is retrieved

            Only data with modification time stamp greater that `date` is retrieved.

            * `datetime.datetime(2020, 11, 27)`: Retrieve data which is is newer than this time stamp
            * 'latest': Retrieve data which is newer than the newest data already in the table.
              .. warning::

                 Don't use 'latest' in combination with `limit`. This might lead to unexpected results.
            * `None`: Complete backfill

            Defaults to `None`.
        limit: int
            Maximum number of locations to download.
            Defaults to `None` which means no limit is set and all available data is queried. Use with care!
        delete_additional_data_requests: bool
            Useful to speed up download of data. Ignores existence of already created requests for additional data and
            skips deletion these.
        """

        # Set limit to a number >> number of locations
        if not limit:
            limit = 10 ** 7

        # Find newest data date if date="latest"
        if date == "latest":
            with session_scope() as session:
                date_queried = (
                    session.query(orm.LocationExtended.DatumLetzteAktualisierung)
                    .order_by(orm.LocationExtended.DatumLetzteAktualisierung.desc())
                    .first()
                )
                if date_queried:
                    date = date_queried[0]
                else:
                    date = None

        locations_basic = self.mastr_dl.basic_location_data(limit, date_from=date)

        for locations_chunk in locations_basic:

            # Remove duplicates returned from API
            locations_chunk_unique = [
                location
                for n, location in enumerate(locations_chunk)
                if location["LokationMastrNummer"]
                not in [_["LokationMastrNummer"] for _ in locations_chunk[n + 1 :]]
            ]
            locations_unique_ids = [
                _["LokationMastrNummer"] for _ in locations_chunk_unique
            ]

            with session_scope() as session:

                # Find units that are already in the DB
                common_ids = [
                    _.LokationMastrNummer
                    for _ in session.query(
                        orm.LocationBasic.LokationMastrNummer
                    ).filter(
                        orm.LocationBasic.LokationMastrNummer.in_(locations_unique_ids)
                    )
                ]

                # Create instances for new data and for updated data
                insert = []
                updated = []
                for location in locations_chunk_unique:
                    # In case data for the unit already exists, only update if new data is newer
                    if location["LokationMastrNummer"] in common_ids:
                        if session.query(
                            exists().where(
                                orm.LocationBasic.LokationMastrNummer
                                == location["LokationMastrNummer"]
                            )
                        ).scalar():
                            updated.append(location)
                            session.merge(orm.LocationBasic(**location))
                    # In case of new data, just insert
                    else:
                        insert.append(location)
                session.bulk_save_objects([orm.LocationBasic(**u) for u in insert])
                inserted_and_updated = insert + updated

                # Create data requests for all newly inserted and updated locations
                new_requests = []
                for location in inserted_and_updated:
                    new_requests.append(
                        {
                            "LokationMastrNummer": location["LokationMastrNummer"],
                            "location_type": self.unit_type_map[
                                location["Lokationtyp"]
                            ],
                            "request_date": datetime.datetime.now(
                                tz=datetime.timezone.utc
                            ),
                        }
                    )

                # Delete old data requests
                if delete_additional_data_requests:
                    ids_to_delete = [
                        _["LokationMastrNummer"] for _ in inserted_and_updated
                    ]
                    session.query(orm.AdditionalLocationsRequested).filter(
                        orm.AdditionalLocationsRequested.LokationMastrNummer.in_(
                            ids_to_delete
                        )
                    ).filter(
                        orm.AdditionalLocationsRequested.request_date
                        < datetime.datetime.now(tz=datetime.timezone.utc)
                    ).delete(
                        synchronize_session="fetch"
                    )
                    session.commit()

                # Do bulk insert of new data requests
                session.bulk_insert_mappings(
                    orm.AdditionalLocationsRequested, new_requests
                )

    def retrieve_additional_data(
        self, technology, data_type, limit=None, chunksize=1000
    ):
        """
        Retrieve additional unit data

        Execute additional data requests stored in :class:`open_mastr.soap_api.orm.AdditionalDataRequested`.
        See also docs of :meth:`open_mastr.soap_api.download.py.MaStRDownload.additional_data` for more
        information on how data is downloaded.

        Parameters
        ----------
        technology: `str` or `list` of `str`
            See list of available technologies in
            :meth:`open_mastr.soap_api.download.py.MaStRDownload.download_power_plants`.
        data_type: `str`
            Select type of additional data that is to be retrieved. Choose from
            "unit_data", "eeg_data", "kwk_data", "permit_data".
        limit: int
            Limit number of units that data is download for. Defaults to `None` which refers
            to query data for existing data requests, for example created by
            :meth:`~.create_additional_data_requests`.
        chunksize: int
            Data is downloaded and inserted into the database in chunks of `chunksize`.
            Defaults to 1000.
        """

        # Mapping of download from MaStRDownload
        download_functions = {
            "unit_data": "extended_unit_data",
            "eeg_data": "eeg_unit_data",
            "kwk_data": "kwk_unit_data",
            "permit_data": "permit_unit_data",
        }

        if not limit:
            limit = 10 ** 8
        if chunksize > limit:
            chunksize = limit

        units_queried = 0
        while units_queried < limit:

            with session_scope() as session:

                requested_chunk = (
                    session.query(orm.AdditionalDataRequested)
                    .filter(
                        and_(
                            orm.AdditionalDataRequested.data_type == data_type,
                            orm.AdditionalDataRequested.technology == technology,
                        )
                    )
                    .limit(chunksize)
                )

                ids = [_.additional_data_id for _ in requested_chunk]

                number_units_merged = 0
                deleted_units = []
                if ids:
                    # Retrieve data
                    unit_data, missed_units = self.mastr_dl.additional_data(
                        technology, ids, download_functions[data_type]
                    )
                    missed_units_ids = [u[0] for u in missed_units]
                    unit_data = flatten_dict(unit_data)

                    # Prepare data and add to database table
                    for unit_dat in unit_data:
                        unit_dat["DatenQuelle"] = "API"
                        unit_dat["DatumDownload"] = date.today()
                        # Remove query status information from response
                        for exclude in [
                            "Ergebniscode",
                            "AufrufVeraltet",
                            "AufrufVersion",
                            "AufrufLebenszeitEnde",
                        ]:
                            del unit_dat[exclude]

                        # Pre-serialize dates/datetimes and decimal in hydro Ertuechtigung
                        # This is required because sqlalchemy does not know how serialize dates/decimal of a JSON
                        if "Ertuechtigung" in unit_dat.keys():
                            for ertuechtigung in unit_dat["Ertuechtigung"]:
                                if ertuechtigung["DatumWiederinbetriebnahme"]:
                                    ertuechtigung[
                                        "DatumWiederinbetriebnahme"
                                    ] = ertuechtigung[
                                        "DatumWiederinbetriebnahme"
                                    ].isoformat()
                                ertuechtigung["ProzentualeErhoehungDesLv"] = float(
                                    ertuechtigung["ProzentualeErhoehungDesLv"]
                                )
                        # The NetzbetreiberMastrNummer is handed over as type:list, hence non-compatible with sqlite)
                        # This replaces the list with the first (string)element in the list to make it sqlite compatible
                        if "NetzbetreiberMastrNummer" in unit_dat.keys():
                            if type(unit_dat["NetzbetreiberMastrNummer"]) == list:
                                if len(unit_dat["NetzbetreiberMastrNummer"]) > 0:
                                    unit_dat["NetzbetreiberMastrNummer"] = unit_dat[
                                        "NetzbetreiberMastrNummer"
                                    ][0]
                                else:
                                    unit_dat["NetzbetreiberMastrNummer"] = None

                        # Create new instance and update potentially existing one
                        unit = getattr(orm, self.orm_map[technology][data_type])(
                            **unit_dat
                        )
                        session.merge(unit)
                        number_units_merged += 1

                    session.commit()
                    # Log units where data retrieval was not successful
                    for missed_unit in missed_units:
                        missed = orm.MissedAdditionalData(
                            additional_data_id=missed_unit[0], reason=missed_unit[1]
                        )
                        session.add(missed)

                    # Remove units from additional data request table if additional data was retrieved
                    for requested_unit in requested_chunk:
                        if requested_unit.additional_data_id not in missed_units_ids:
                            session.delete(requested_unit)
                            deleted_units.append(requested_unit.additional_data_id)

                    # Update while iteration condition
                    units_queried += len(ids)

                    log.info(
                        f"Downloaded data for {len(unit_data)} units ({len(ids)} requested). "
                        f"Missed units: {len(missed_units)}. Deleted requests: {len(deleted_units)}."
                    )

            # Emergency break out: if now new data gets inserted/update, don't retrieve any further data
            if number_units_merged == 0:
                log.info("No further data is requested")
                break

    def retrieve_additional_location_data(
        self, location_type, limit=None, chunksize=1000
    ):
        """
        Retrieve extended location data

        Execute additional data requests stored in :class:`open_mastr.soap_api.orm.AdditionalLocationsRequested`.
        See also docs of :meth:`open_mastr.soap_api.download.py.MaStRDownload.additional_data` for more
        information on how data is downloaded.

        Parameters
        ----------
        location_type: `str`
            Select type of location that is to be retrieved. Choose from
            "location_elec_generation", "location_elec_consumption", "location_gas_generation",
            "location_gas_consumption".
        limit: int
            Limit number of locations that data is download for. Defaults to `None` which refers
            to query data for existing data requests.
        chunksize: int
            Data is downloaded and inserted into the database in chunks of `chunksize`.
            Defaults to 1000.
        """

        # Process arguments
        if not limit:
            limit = 10 ** 8
        if chunksize > limit:
            chunksize = limit

        locations_queried = 0
        while locations_queried < limit:

            with session_scope() as session:
                # Get a chunk
                requested_chunk = (
                    session.query(orm.AdditionalLocationsRequested)
                    .filter(
                        orm.AdditionalLocationsRequested.location_type == location_type
                    )
                    .limit(chunksize)
                )
                ids = [_.LokationMastrNummer for _ in requested_chunk]

                # Reset number of locations inserted or updated for this chunk
                number_locations_merged = 0
                deleted_locations = []
                if ids:
                    # Retrieve data
                    location_data, missed_locations = self.mastr_dl.additional_data(
                        location_type, ids, "location_data"
                    )
                    missed_locations_ids = [loc[0] for loc in missed_locations]

                    # Prepare data and add to database table
                    for location_dat in location_data:
                        # Remove query status information from response
                        for exclude in [
                            "Ergebniscode",
                            "AufrufVeraltet",
                            "AufrufVersion",
                            "AufrufLebenszeitEnde",
                        ]:
                            del location_dat[exclude]

                        # Make data types JSON serializable
                        location_dat["DatumLetzteAktualisierung"] = location_dat[
                            "DatumLetzteAktualisierung"
                        ].isoformat()
                        for grid_connection in location_dat["Netzanschlusspunkte"]:
                            grid_connection["letzteAenderung"] = grid_connection[
                                "letzteAenderung"
                            ].isoformat()
                            for field_name in [
                                "MaximaleEinspeiseleistung",
                                "MaximaleAusspeiseleistung",
                                "Nettoengpassleistung",
                                "Netzanschlusskapazitaet",
                            ]:
                                if field_name in grid_connection.keys():
                                    if grid_connection[field_name]:
                                        grid_connection[field_name] = float(
                                            grid_connection[field_name]
                                        )
                        # This converts dates of type:string to type:datetime to match column data types in orm.py
                        if type(location_dat["DatumLetzteAktualisierung"]) == str:
                            location_dat[
                                "DatumLetzteAktualisierung"
                            ] = datetime.datetime.strptime(
                                location_dat["DatumLetzteAktualisierung"],
                                "%Y-%m-%dT%H:%M:%S.%f",
                            )

                        # Create new instance and update potentially existing one
                        location = orm.LocationExtended(**location_dat)
                        session.merge(location)
                        number_locations_merged += 1

                    session.commit()
                    # Log locations where data retrieval was not successful
                    for missed_location in missed_locations:
                        missed = orm.MissedExtendedLocation(
                            LokationMastrNummer=missed_location[0],
                            reason=missed_location[1],
                        )
                        session.add(missed)

                    # Remove locations from additional data request table if additional data was retrieved
                    for requested_location in requested_chunk:
                        if (
                            requested_location.LokationMastrNummer
                            not in missed_locations_ids
                        ):
                            session.delete(requested_location)
                            deleted_locations.append(
                                requested_location.LokationMastrNummer
                            )

                    # Update while iteration condition
                    locations_queried += len(ids)

                    log.info(
                        f"Downloaded data for {len(location_data)} locations ({len(ids)} requested). "
                        f"Missed locations: {len(missed_locations_ids)}. Deleted requests: "
                        f"{len(deleted_locations)}."
                    )

            # Emergency break out: if now new data gets inserted/update, don't retrieve any further data
            if number_locations_merged == 0:
                log.info("No further data is requested")
                break

    def create_additional_data_requests(
        self,
        technology,
        data_types=["unit_data", "eeg_data", "kwk_data", "permit_data"],
        delete_existing=True,
    ):
        """
        Create new requests for additional unit data

        For units that exist in basic_units but not in the table for additional data of `data_type`, a new data request
        is submitted.

        Parameters
        ----------
        technology: str
            Specify technology additional data should be requested for.
        data_types: list
            Select type of additional data that is to be requested. Defaults to all data that is available for a
            technology.
        delete_existing: bool
            Toggle deletion of already existing requests for additional data.
            Defaults to True.
        """

        data_requests = []

        with session_scope() as session:
            # Check which additional data is missing
            for data_type in data_types:
                data_type_available = self.orm_map[technology].get(data_type, None)

                # Only proceed if this data type is available for this technology
                if data_type_available:
                    log.info(
                        f"Create requests for additional data of type {data_type} for {technology}"
                    )

                    # Get ORM for additional data by technology and data_type
                    additional_data_orm = getattr(orm, data_type_available)

                    # Delete prior additional data requests for this technology and data_type
                    if delete_existing:
                        session.query(orm.AdditionalDataRequested).filter(
                            orm.AdditionalDataRequested.technology == technology,
                            orm.AdditionalDataRequested.data_type == data_type,
                        ).delete()
                        session.commit()

                    # Query database for missing additional data
                    if data_type == "unit_data":
                        units_for_request = (
                            session.query(orm.BasicUnit)
                            .outerjoin(
                                additional_data_orm,
                                orm.BasicUnit.EinheitMastrNummer
                                == additional_data_orm.EinheitMastrNummer,
                            )
                            .filter(
                                orm.BasicUnit.Einheittyp
                                == self.unit_type_map_reversed[technology]
                            )
                            .filter(additional_data_orm.EinheitMastrNummer.is_(None))
                            .filter(orm.BasicUnit.EinheitMastrNummer.isnot(None))
                        )
                    elif data_type == "eeg_data":
                        units_for_request = (
                            session.query(orm.BasicUnit)
                            .outerjoin(
                                additional_data_orm,
                                orm.BasicUnit.EegMastrNummer
                                == additional_data_orm.EegMastrNummer,
                            )
                            .filter(
                                orm.BasicUnit.Einheittyp
                                == self.unit_type_map_reversed[technology]
                            )
                            .filter(additional_data_orm.EegMastrNummer.is_(None))
                            .filter(orm.BasicUnit.EegMastrNummer.isnot(None))
                        )
                    elif data_type == "kwk_data":
                        units_for_request = (
                            session.query(orm.BasicUnit)
                            .outerjoin(
                                additional_data_orm,
                                orm.BasicUnit.KwkMastrNummer
                                == additional_data_orm.KwkMastrNummer,
                            )
                            .filter(
                                orm.BasicUnit.Einheittyp
                                == self.unit_type_map_reversed[technology]
                            )
                            .filter(additional_data_orm.KwkMastrNummer.is_(None))
                            .filter(orm.BasicUnit.KwkMastrNummer.isnot(None))
                        )
                    elif data_type == "permit_data":
                        units_for_request = (
                            session.query(orm.BasicUnit)
                            .outerjoin(
                                additional_data_orm,
                                orm.BasicUnit.GenMastrNummer
                                == additional_data_orm.GenMastrNummer,
                            )
                            .filter(
                                orm.BasicUnit.Einheittyp
                                == self.unit_type_map_reversed[technology]
                            )
                            .filter(additional_data_orm.GenMastrNummer.is_(None))
                            .filter(orm.BasicUnit.GenMastrNummer.isnot(None))
                        )
                    else:
                        raise ValueError(
                            f"Data type {data_type} is not a valid option."
                        )

                    # Prepare data for additional data request
                    for basic_unit in units_for_request:
                        data_request = {
                            "EinheitMastrNummer": basic_unit.EinheitMastrNummer,
                            "technology": self.unit_type_map[basic_unit.Einheittyp],
                            "data_type": data_type,
                            "request_date": datetime.datetime.now(
                                tz=datetime.timezone.utc
                            ),
                        }
                        if data_type == "unit_data":
                            data_request[
                                "additional_data_id"
                            ] = basic_unit.EinheitMastrNummer
                        elif data_type == "eeg_data":
                            data_request[
                                "additional_data_id"
                            ] = basic_unit.EegMastrNummer
                        elif data_type == "kwk_data":
                            data_request[
                                "additional_data_id"
                            ] = basic_unit.KwkMastrNummer
                        elif data_type == "permit_data":
                            data_request[
                                "additional_data_id"
                            ] = basic_unit.GenMastrNummer
                        data_requests.append(data_request)

            # Insert new requests for additional data into database
            session.bulk_insert_mappings(orm.AdditionalDataRequested, data_requests)

    def dump(self, dumpfile="open-mastr-continuous-update.backup"):
        """
        Dump MaStR database.

        Parameters
        ----------
        dumpfile : str or path-like, optional
            Save path for dump including filename. When only a filename is given, the dump is saved to CWD.
        """
        dump_cmd = (
            f"pg_dump -Fc "
            f"-f {dumpfile} "
            f"-n mastr_mirrored "
            f"-h localhost "
            f"-U open-mastr "
            f"-p 55443 "
            f"open-mastr"
        )

        proc = subprocess.Popen(dump_cmd, shell=True, env={"PGPASSWORD": "open-mastr"})
        proc.wait()

    def restore(self, dumpfile):
        """
        Restore the MaStR database from an SQL dump.

        Parameters
        ----------
        dumpfile : str or path-like, optional
            Save path for dump including filename. When only a filename is given, the dump is restored from CWD.


        Warnings
        --------
        If tables that are restored from the dump contain data, restore doesn't work!

        """
        # Interpret file name and path
        dump_file_dir, dump_file = os.path.split(dumpfile)
        cwd = os.path.abspath(os.path.dirname(dump_file_dir))

        # Define import of SQL dump with pg_restore
        restore_cmd = (
            f"pg_restore -h localhost -U open-mastr -p 55443 -d open-mastr {dump_file}"
        )
        restore_cmd = shlex.split(restore_cmd)

        # Execute restore command
        proc = subprocess.Popen(
            restore_cmd,
            shell=False,
            env={"PGPASSWORD": "open-mastr"},
            cwd=cwd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        proc.wait()

    def to_csv(
        self,
        technology=None,
        limit=None,
        additional_data=["unit_data", "eeg_data", "kwk_data", "permit_data"],
        statistic_flag="B",
    ):
        """
        Export a snapshot MaStR data from mirrored database to CSV

        During the export, additional available data is joined on list of basic units. A CSV file for each technology is
        created separately because of multiple non-overlapping columns.
        Duplicate columns for a single technology (a results on data from different sources) are suffixed.

        The data in the database probably has duplicates because of the history how data was collected in the
        Marktstammdatenregister. Consider to use the parameter `statistic_flag`. Read more in the
        `documentation <https://www.marktstammdatenregister.de/MaStRHilfe/subpages/statistik.html>`_ of the original
        data source.

        Along with the CSV files, metadata is saved in the file `datapackage.json`.

        Parameters
        ----------
        technology: `str` or `list` of `str`
            See list of available technologies in
            :meth:`open_mastr.soap_api.download.py.MaStRDownload.download_power_plants`
        limit: int
            Limit number of rows
        additional_data: `list`
            Defaults to "export all available additional data" which is described by
            `["unit_data", "eeg_data", "kwk_data", "permit_data"]`.
        statistic_flag: `str`
            Choose between 'A' or 'B' (default) to select a subset of the data for the export to CSV.

            * 'B': Migrated that was migrated to the Martstammdatenregister + newly registered units with commissioning
              date after 31.01.2019 (recommended for statistical purposes).
            * 'A':  Newly registered units with commissioning date before 31.01.2019
            * None: Export all data
        """

        create_data_dir()

        # Make sure input in either str or list
        if isinstance(technology, str):
            technology = [technology]
        elif not isinstance(technology, (list, None)):
            raise TypeError("Parameter technology must be of type `str` or `list`")

        renaming = column_renaming()

        with session_scope() as session:

            for tech in technology:
                unit_data_orm = getattr(orm, self.orm_map[tech]["unit_data"], None)
                eeg_data_orm = getattr(
                    orm, self.orm_map[tech].get("eeg_data", "KeyNotAvailable"), None
                )
                kwk_data_orm = getattr(
                    orm, self.orm_map[tech].get("kwk_data", "KeyNotAvailable"), None
                )
                permit_data_orm = getattr(
                    orm, self.orm_map[tech].get("permit_data", "KeyNotAvailable"), None
                )

                # Define query based on available tables for tech and user input
                subtables = partially_suffixed_columns(
                    orm.BasicUnit,
                    renaming["basic_data"]["columns"],
                    renaming["basic_data"]["suffix"],
                )
                if unit_data_orm and "unit_data" in additional_data:
                    subtables.extend(
                        partially_suffixed_columns(
                            unit_data_orm,
                            renaming["unit_data"]["columns"],
                            renaming["unit_data"]["suffix"],
                        )
                    )
                if eeg_data_orm and "eeg_data" in additional_data:
                    subtables.extend(
                        partially_suffixed_columns(
                            eeg_data_orm,
                            renaming["eeg_data"]["columns"],
                            renaming["eeg_data"]["suffix"],
                        )
                    )
                if kwk_data_orm and "kwk_data" in additional_data:
                    subtables.extend(
                        partially_suffixed_columns(
                            kwk_data_orm,
                            renaming["kwk_data"]["columns"],
                            renaming["kwk_data"]["suffix"],
                        )
                    )
                if permit_data_orm and "permit_data" in additional_data:
                    subtables.extend(
                        partially_suffixed_columns(
                            permit_data_orm,
                            renaming["permit_data"]["columns"],
                            renaming["permit_data"]["suffix"],
                        )
                    )
                query = Query(subtables, session=session)

                # Define joins based on available tables for tech and user input
                if unit_data_orm and "unit_data" in additional_data:
                    query = query.join(
                        unit_data_orm,
                        orm.BasicUnit.EinheitMastrNummer
                        == unit_data_orm.EinheitMastrNummer,
                        isouter=True,
                    )
                if eeg_data_orm and "eeg_data" in additional_data:
                    query = query.join(
                        eeg_data_orm,
                        orm.BasicUnit.EegMastrNummer == eeg_data_orm.EegMastrNummer,
                        isouter=True,
                    )
                if kwk_data_orm and "kwk_data" in additional_data:
                    query = query.join(
                        kwk_data_orm,
                        orm.BasicUnit.KwkMastrNummer == kwk_data_orm.KwkMastrNummer,
                        isouter=True,
                    )
                if permit_data_orm and "permit_data" in additional_data:
                    query = query.join(
                        permit_data_orm,
                        orm.BasicUnit.GenMastrNummer == permit_data_orm.GenMastrNummer,
                        isouter=True,
                    )

                # Restricted to technology
                query = query.filter(
                    orm.BasicUnit.Einheittyp == self.unit_type_map_reversed[tech]
                )

                # Decide if migrated data or data of newly registered units or both is selected
                if statistic_flag and "unit_data" in additional_data:
                    query = query.filter(unit_data_orm.StatisikFlag == statistic_flag)

                # Limit returned rows of query
                if limit:
                    query = query.limit(limit)

                # Read data into pandas.DataFrame
                df = pd.read_sql(
                    query.statement, query.session.bind, index_col="EinheitMastrNummer"
                )

                # Remove newline statements from certain strings
                for col in ["Aktenzeichen", "Behoerde"]:
                    df[col] = df[col].str.replace("\r", "")

                # Make sure no duplicate column names exist
                assert not any(df.columns.duplicated())

                # Save to CSV
                to_csv(df, tech)

            # Create and save data package metadata file along with data
            data_path = get_data_version_dir()
            filenames = get_filenames()
            metadata_file = os.path.join(data_path, filenames["metadata"])

            mastr_technologies = [
                self.unit_type_map_reversed[tech] for tech in technology
            ]
            newest_date = (
                session.query(orm.BasicUnit.DatumLetzteAktualisierung)
                .filter(orm.BasicUnit.Einheittyp.in_(mastr_technologies))
                .order_by(orm.BasicUnit.DatumLetzteAktualisierung.desc())
                .first()[0]
            )
        metadata = datapackage_meta_json(newest_date, technology, json_serialize=False)

        with open(metadata_file, "w", encoding="utf-8") as f:
            json.dump(metadata, f, ensure_ascii=False, indent=4)

    def reverse_fill_basic_units(self):
        """
        The basic_units table is empty after bulk download. To enable csv export, the table is filled from extended
        tables reversely.
        .. warning::
        The basic_units table will be dropped and then recreated.
        Returns -------

        """

        technology = [
            "solar",
            "wind",
            "biomass",
            "combustion",
            "gsgk",
            "hydro",
            "nuclear",
            "storage",
        ]

        with session_scope() as session:
            # Empty the basic_units table, because it will be filled entirely from extended tables
            session.query(getattr(orm, "BasicUnit", None)).delete()

            for tech in technology:
                # Get the class of extended table
                unit_data_orm = getattr(orm, self.orm_map[tech]["unit_data"], None)
                basic_unit_column_names = [
                    column.name
                    for column in getattr(orm, "BasicUnit", None).__mapper__.columns
                ]

                unit_columns_to_reverse_fill = [
                    column
                    for column in unit_data_orm.__mapper__.columns
                    if column.name in basic_unit_column_names
                ]
                unit_column_names_to_reverse_fill = [
                    column.name for column in unit_columns_to_reverse_fill
                ]

                # Add Einheittyp artificially
                unit_typ = "'" + self.unit_type_map_reversed.get(tech, None) + "'"
                unit_columns_to_reverse_fill.append(
                    literal_column(unit_typ).label("Einheittyp")
                )
                unit_column_names_to_reverse_fill.append("Einheittyp")

                # Build query
                query = Query(unit_columns_to_reverse_fill, session=session)
                insert_query = insert(orm.BasicUnit).from_select(
                    unit_column_names_to_reverse_fill, query
                )

                session.execute(insert_query)

    def locations_to_csv(self, location_type, limit=None):
        """
        Save location raw data to CSV file

        During data export to CSV file, data is reshaped to tabular format. Data stored in JSON types is flattened and
        concated to separate rows.

        Parameters
        ----------
        location_type: `str`
            Select type of location that is to be retrieved. Choose from
            "location_elec_generation", "location_elec_consumption", "location_gas_generation",
            "location_gas_consumption".
        limit: int
            Limit number of rows. Defaults to None which implies all rows are selected.
        """
        location_type_shorthand = {
            "location_elec_generation": "SEL",
            "location_elec_consumption": "SVL",
            "location_gas_generation": "GEL",
            "location_gas_consumption": "GVL",
        }

        with session_scope() as session:
            # Load basic and extended location data into DataFrame
            locations_extended = (
                session.query(
                    *[
                        c
                        for c in orm.LocationBasic.__table__.columns
                        if c.name not in ["NameDerTechnischenLokation"]
                    ],
                    *[
                        c
                        for c in orm.LocationExtended.__table__.columns
                        if c.name not in ["MaStRNummer"]
                    ],
                )
                .join(
                    orm.LocationExtended,
                    orm.LocationBasic.LokationMastrNummer
                    == orm.LocationExtended.MastrNummer,
                )
                .filter(
                    orm.LocationBasic.LokationMastrNummer.startswith(
                        location_type_shorthand[location_type]
                    )
                )
                .limit(limit)
            )

            df = pd.read_sql(locations_extended.statement, session.bind)

        # Expand data about grid connection points from dict into separate columns
        df_expanded = pd.concat(
            [pd.DataFrame(x) for x in df["Netzanschlusspunkte"]], keys=df.index
        ).reset_index(level=1, drop=True)
        df = (
            df.drop("Netzanschlusspunkte", axis=1)
            .join(df_expanded)
            .reset_index(drop=True)
        )

        # Expand data about related units into separate columns (with lists of related units)
        df = df.drop("VerknuepfteEinheiten", axis=1).join(
            df["VerknuepfteEinheiten"].apply(list_of_dicts_to_columns)
        )

        # Save to file
        create_data_dir()
        data_path = get_data_version_dir()
        filenames = get_filenames()
        csv_file = os.path.join(data_path, filenames["raw"][location_type])
        df.to_csv(csv_file, index=False, encoding="utf-8")


def list_of_dicts_to_columns(row):
    """
    Expand data stored in dict to spearate columns

    Parameters
    ----------
    row: list of dict
        Usually apllied using apply on a column of a pandas DataFrame, hence, a Series. This column of the
        DataFrame should comprise of a single-level dict with an arbitrary number of columns. Each key is
        transformed into a new column, while data from each dict inside the list is concatenated by key. Such
        that the data is stored into a list for each key/column.

    Returns
    -------
    pd.Series
        Pandas Series with keys as columns and values concatenated to a list for each key.
    """
    columns = {k: [] for dic in row for k, _ in dic.items()}
    for dic in row:
        for k, v in dic.items():
            columns[k].append(v)

    new_cols_df = pd.Series(columns)
    return new_cols_df


def partially_suffixed_columns(mapper, column_names, suffix):
    """
    Add a suffix to a subset of ORM map tables for a query

    Parameters
    ----------
    mapper:
        SQLAlchemy ORM table mapper
    column_names: list
        Names of columns to be suffixed
    suffix: str
        Suffix that is append like + "_" + suffix

    Returns
    -------
    list
        List of ORM table mapper instance
    """
    columns = [_ for _ in mapper.__mapper__.columns]
    columns_renamed = [
        _.label(f"{_.name}_{suffix}") if _.name in column_names else _ for _ in columns
    ]
    return columns_renamed



import os

from open_mastr.soap_api.metadata.create import column_docs_csv


def generate_data_docs():
    technologies = [
        "solar",
        "wind",
        "biomass",
        "combustion",
        "gsgk",
        "hydro",
        "nuclear",
        "storage",
    ]
    raw_data_doc_files = column_docs_csv(technologies, "_data/raw/")

    raw_data_string = "Raw data\n========\n\nRaw data retrieved from MaStR database is structured as follows\n\n"

    for tech, data_table_doc in zip(technologies, raw_data_doc_files):
        section = f"{tech}\n-------\n\n"

        csv_include = (
            f".. csv-table::\n"
            f"   :file: {os.path.join('raw', data_table_doc)}\n"
            "   :widths: 20, 35, 15, 15\n"
            "   :header-rows: 1\n\n\n"
        )

        raw_data_string += section + csv_include
    with open("_data/raw_data.rst", "w") as raw_data_fh:
        raw_data_fh.write(raw_data_string)



from setuptools import setup, find_packages
from os import path
import os

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="open_mastr",
    packages=[
        "open_mastr",
        "open_mastr.soap_api",
        "open_mastr.soap_api.config",
        "open_mastr.soap_api.metadata",
        "open_mastr.utils",
        "open_mastr.xml_download",
    ],
    version="0.11.1",
    description="The python package open_mastr provides an interface for "
                "accessing the data. It contains methods to download and "
                "parse the xml files (bulk) and the web service (API).",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/OpenEnergyPlatform/open-MaStR",
    download_url="https://github.com/OpenEnergyPlatform/open-MaStR/archive"
                 "/refs/tags/v0.11.1.tar.gz",
    maintainer_email="datenzentrum@rl-institut.de",

    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires=">=3.8, <4",
    install_requires=[
        "pandas>=1.4",
        "numpy",
        "sqlalchemy",
        "psycopg2-binary",
        "zeep",
        "tqdm",
        "requests",
        "keyring",
        "pynodo",
        "tqdm",
        "beautifulsoup4",
        "pyyaml",
        "xmltodict",
    ],
    extras_require={
        "dev": [
            "flake8",
            "pylint",
            "pytest",
            "pytest-dependency",
            "xmltodict",  # TODO: Only needed in metadata
        ]
    },
    package_data={
        "open_mastr": [
            os.path.join("soap_api", "config", "*.yml"),
            os.path.join("soap_api", "metadata", "LICENSE"),
        ]
    },
    project_urls={
        "Bug Reports": "https://github.com/OpenEnergyPlatform/open-MaStR/issues",
        "Source": "https://github.com/OpenEnergyPlatform/open-MaStR",
    },
)

#os.system("ulimit -n 1000000")



import os
from open_mastr.soap_api.config import get_project_home_dir

def create_credentials_file():
    """Use token and user stored in GitHub secrets for creating credentials file

    This is used to allow test workflow to access MaStR database.
    """
    credentials_file = os.path.join(get_project_home_dir(), 'config', 'credentials.cfg')

    token = os.getenv("MASTR_TOKEN")
    user = os.getenv("MASTR_USER")
    section_title = "[MaStR]"

    file_content = f"{section_title}\n" \
        f"user = {user}\n" \
        f"token = {token}\n"

    with open(credentials_file, "w") as credentials_fh:
        credentials_fh.write(file_content)



"""
Read and write user config file for data base credentials

The config file storing the MaStR user name is located in
`~/.open-MaStR/config.ini`.

SPDX-License-Identifier: AGPL-3.0-or-later
"""

__copyright__ = "© Reiner Lemoine Institut"
__license__ = "GNU Affero General Public License Version 3 (AGPL-3.0)"
__url__ = "https://www.gnu.org/licenses/agpl-3.0.en.html"
__author__ = "gplssm"
__issue__ = "https://github.com/OpenEnergyPlatform/examples/issues/83"
__version__ = "v0.10.0"

import os
import configparser as cp
from open_mastr.soap_api.config import get_project_home_dir
import keyring

import logging
log = logging.getLogger(__name__)


def _load_config_file():

    config_file = os.path.join(get_project_home_dir(), 'config', 'credentials.cfg')
    cfg = cp.ConfigParser()

    # if not os.path.isdir(open_mastr_home):
    #     create_open_mastr_dir(open_mastr_home)

    if os.path.isfile(config_file):
        cfg.read(config_file)
        return cfg
    else:
        with open(config_file, 'w') as configfile:
            cfg.write(configfile)
        return cfg


def get_mastr_user():
    """
    Read MaStR user name (MaStR Nummer/ MaStR Akteursnummer) from config file

    If no user name is available or the config file doesn't exist, it's
    created ad-hoc and the user is prompted for input.

    Returns
    -------
    str : MaStR user name (MaStR Nummer/ MaStR Akteursnummer)
    """
    cfg = _load_config_file()
    section = "MaStR"
    cfg_path = os.path.join(get_project_home_dir(), 'config', 'credentials.cfg')

    try:
        user = cfg.get(section, "user")
        return user
    except (cp.NoSectionError, cp.NoOptionError):
        # Actually, an error should be raised here. But this conflicts
        # with automatic calls from MaStRDownloadFactory
        # except cp.NoSectionError:
        #     raise cp.Error(f"Section {section} not found in {cfg_path}")
        # except cp.NoOptionError:
        #     raise cp.Error(f"The option 'user' could not by found in the section "
        #                    f"{section} in file {cfg_path}.")
        log.warning(f"The option 'user' could not by found in the section "
                    f"{section} in file {cfg_path}. "
                    f"You might run into trouble when downloading data via the MaStR API."
                    f"\n Bulk download works without option 'user'.")
        return None


def check_and_set_mastr_user():
    """Checks if MaStR user is stored, otherwise asks for it."""

    user = get_mastr_user()

    if not user:
        credentials_file = os.path.join(get_project_home_dir(), 'config', 'credentials.cfg')
        cfg = _load_config_file()

        user = input('\n\nCannot not find a MaStR user name in {config_file}.\n\n'
                     'Please enter MaStR-ID (pattern: SOM123456789012): '
                     ''.format(config_file=credentials_file))
        cfg["MaStR"] = {"user": user}

        with open(credentials_file, 'w') as configfile:
            cfg.write(configfile)

    return user


def get_mastr_token(user):
    """
    Get MaStR token of an associated role from keyring

    If the token/password is not available in the keyring, the user is asked
    to provide the token which is then stored in the keyring

    Parameters
    ----------
    user : str
        MaStR user name (MaStR Nummer/ MaStR Akteursnummer)

    Returns
    -------
    str : Token (password)
    """

    # Try to get password from keyring
    keyring.get_keyring()
    # Retrieving password from keyring does currently fail on headless systems
    # Prevent from breaking program execution with following try/except clause
    section = "MaStR"
    cfg_path = os.path.join(get_project_home_dir(), 'config', 'credentials.cfg')
    try:
        password = keyring.get_password(section, user)
    except:
        password = None

    # No password stored in keyring, try config file
    if not password:
        cfg = _load_config_file()
        try:
            password = cfg.get(section, "token")
        except (cp.NoSectionError, cp.NoOptionError):
            log.warning(f"The option 'token' could not by found in the section "
                        f"{section} in file {cfg_path}. "
                        f"You might run into trouble when downloading data via the MaStR API."
                        f"\n Bulk download works without option 'token'.")
            password = None
    return password


def check_and_set_mastr_token(user):
    """Checks if MaStR token is stored, otherwise asks for it."""

    password = get_mastr_token(user)

    if not password:
        cfg = _load_config_file()
        credentials_file = os.path.join(get_project_home_dir(), 'config', 'credentials.cfg')

        # If also no password in credentials file, ask the user to input password
        # Two options: (1) storing in keyring; (2) storing in config file
        password = input('\n\nCannot not find a MaStR password, neither in keyring nor in {config_file}.\n\n'
                         "Please enter a valid access token of a role (Benutzerrolle) "
                         "associated to the user {user}.\n"
                         "The token might look like: "
                         "koo5eixeiQuoi'w8deighai8ahsh1Ha3eib3coqu7ceeg%ies...\n".format(
            config_file=credentials_file,
            user=user))

        # let the user decide where to store the password
        # (1) keyring
        # (2) credentials.cfg
        # (0) don't store, abort
        # Wait for correct input
        while True:
            choice = int(input("Where do you want to store your password?\n"
                            "\t(1) Keyring (default, hit ENTER to select)\n"
                            "\t(2) Config file (credendials.cfg)\n"
                            "\t(0) Abort. Don't store password\n") or "1\n")
            # check if choice is valid input
            if choice in [0, 1, 2]:
                break

        # Do action according to input
        if choice == 0:
            pass
        elif choice == 1:
            keyring.set_password("MaStR", user, password)
        elif choice == 2:
            cfg["MaStR"] = {"user": user, "token": password}
            with open(credentials_file, 'w') as configfile:
                cfg.write(configfile)
        else:
            log.error("No clue what happened here!?")

    return password


def get_zenodo_token():
    """
    Read Zenodo token from config file (credentials.cfg)

    Returns
    -------
    str
        Zenodo token
    """
    cfg = _load_config_file()
    section = "Zenodo"

    try:
        user = cfg.get(section, "token")
        return user
    except (cp.NoSectionError, cp.NoOptionError):
        return None